from __future__ import print_function

import argparse
import json
import random

import os

import time

from generator import import_synthetic_data_files
from inc_logging import LoggingObserver
from incremental_learner import AllViolationsStrategy, RandomViolationsStrategy, WeightedRandomViolationsStrategy, \
    MaxViolationsStrategy
from k_cnf_smt_learner import KCnfSmtLearner
from k_dnf_smt_learner import KDnfSmtLearner
from parameter_free_learner import learn_bottom_up
from timeout import timeout


class IncrementalConfig(object):
    def __init__(self, initial, initial_size, selection, selection_size):
        self.initial = initial
        self.initial_size = initial_size
        self.selection = selection
        self.selection_size = selection_size
        self.domain = None
        self.data = None
        self.dt_weights = None

    def set_data(self, data):
        self.data = data
        self.dt_weights = None

    def get_dt_weights(self):
        if self.dt_weights is None:
            import dt_selection
            self.dt_weights = [min(d.values()) for d in dt_selection.get_distances(self.domain, self.data)]
        return self.dt_weights

    def get_initial_indices(self):
        if self.initial is None:
            return list(range(len(self.data)))
        elif self.initial == "random":
            return random.sample(range(len(self.data)), self.initial_size)
        elif self.initial == "dt_weighted":
            import sampling
            return sampling.sample_weighted(zip(range(len(self.data)), self.get_dt_weights()), self.initial_size)
        else:
            raise RuntimeError("Unknown initial type {}".format(self.initial))

    def get_selection_strategy(self):
        if self.selection is None:
            return RandomViolationsStrategy(0)
        elif self.selection == "random":
            return RandomViolationsStrategy(self.selection_size)
        elif self.selection == "dt_weighted":
            return WeightedRandomViolationsStrategy(self.selection_size, self.get_dt_weights())
        elif self.selection == "dt":
            return MaxViolationsStrategy(self.selection_size, self.get_dt_weights())
        else:
            raise RuntimeError("Unknown selection type {}".format(self.selection))


def learn_synthetic(input_dir, prefix, results_dir, bias, incremental_config, plot=None, sample_count=None,
                    time_out=None, parameter_free=False):

    input_dir = os.path.abspath(input_dir)
    data_sets = list(import_synthetic_data_files(input_dir, prefix))

    if not os.path.exists(results_dir):
        os.makedirs(results_dir)
    overview = os.path.join(results_dir, "problems.txt")

    if not os.path.isfile(overview):
        flat = {}
    else:
        with open(overview, "r") as f:
            flat = json.load(f)

    for data_set in data_sets:
        synthetic_problem = data_set.synthetic_problem
        data = data_set.samples
        name = synthetic_problem.theory_problem.name
        domain = synthetic_problem.theory_problem.domain

        if name not in flat:
            flat[name] = {}

        print(name)

        seed = hash(time.time())
        random.seed(seed)

        if sample_count is not None and sample_count < len(data):
            data = random.sample(data, sample_count)
        else:
            sample_count = len(data)

        incremental_config.set_data(data)
        incremental_config.domain = domain

        if not parameter_free:
            initial_indices = incremental_config.get_initial_indices()
            h = synthetic_problem.half_space_count
            k = synthetic_problem.formula_count

            if bias == "cnf" or bias == "dnf":
                selection_strategy = incremental_config.get_selection_strategy()
                if bias == "cnf":
                    learner = KCnfSmtLearner(k, h, selection_strategy)
                elif bias == "dnf":
                    learner = KDnfSmtLearner(k, h, selection_strategy)

                if plot is not None and plot and synthetic_problem.bool_count == 0 and synthetic_problem.real_count == 2:
                    import plotting
                    feats = domain.real_vars
                    plots_dir = os.path.join(results_dir, name)
                    exp_id = "{}_{}_{}".format(learner.name, sample_count, seed)
                    learner.add_observer(plotting.PlottingObserver(data, plots_dir, exp_id, *feats))
                log_file = "{}_{}_{}_{}_{}.learning_log.txt".format(name, sample_count, seed, k, h)
                learner.add_observer(LoggingObserver(os.path.join(results_dir, log_file), seed, True, selection_strategy))
            else:
                raise RuntimeError("Unknown bias {}".format(bias))

            result = timeout(learner.learn, [domain, data, initial_indices], duration=time_out)
        else:
            def learn_f(_data, _k, _h):
                selection_strategy = incremental_config.get_selection_strategy()
                if bias == "cnf":
                    learner = KCnfSmtLearner(_k, _h, selection_strategy)
                elif bias == "dnf":
                    learner = KDnfSmtLearner(_k, _h, selection_strategy)
                initial_indices = incremental_config.get_initial_indices()
                log_file = "{}_{}_{}_{}_{}.learning_log.txt".format(name, sample_count, seed, _k, _h)
                learner.add_observer(LoggingObserver(os.path.join(results_dir, log_file), seed, True, selection_strategy))
                return learner.learn(domain, data, initial_indices)

            result, k, h = learn_bottom_up(data, learn_f, 3, 1)
        if result is None:
            flat[name][sample_count] = {"k": k, "h": h, "seed": seed, "bias": bias, "time_out": True}
        else:
            flat[name][sample_count] = {"k": k, "h": h, "seed": seed, "bias": bias, "time_out": False}
        if time_out is not None:
            flat[name][sample_count]["time_limit"] = time_out

    with open(overview, "w") as f:
        json.dump(flat, f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_dir")
    parser.add_argument("prefix")
    parser.add_argument("output_dir")
    parser.add_argument("--bias", default="cnf")
    parser.add_argument("--initial", default="random")
    parser.add_argument("--initial_size", default=20, type=int)
    parser.add_argument("--selection", default="random")
    parser.add_argument("--selection_size", default=10, type=int)
    parser.add_argument("-p", "--plot", action="store_true")
    parser.add_argument("-s", "--samples", default=None, type=int)
    parser.add_argument("-t", "--time_out", default=None, type=int)
    parser.add_argument("-a", "--non_incremental", default=False, action="store_true")
    parser.add_argument("-f", "--parameter_free", default=False, action="store_true")
    parsed = parser.parse_args()

    if parsed.non_incremental:
        inc_config = IncrementalConfig(None, None, None, None)
    else:
        inc_config = IncrementalConfig(parsed.initial, parsed.initial_size, parsed.selection, parsed.selection_size)

    learn_synthetic(parsed.input_dir, parsed.prefix, parsed.output_dir, parsed.bias, inc_config,
                    parsed.plot, parsed.samples, parsed.time_out, parsed.parameter_free)



import glob
import os
import random
import warnings

from .prepare import get_synthetic_db
from incal.util.options import Options, Experiment

from incal.learn import LearnOptions, LearnResults
from incal.util.parallel import run_commands
from .prepare import select_benchmark_files, benchmark_filter, get_benchmark_results_dir


def get_bound_volume(bounds):
    size = 1
    for ub_lb in bounds.values():
        size *= ub_lb[1] - ub_lb[0]
    return size


def rel_ratio(ratio):
    return abs(0.5 - ratio)


def learn_synthetic(input_directory, output_directory, runs, sample_size, processes, time_out, learn_options: LearnOptions):
    commands = []

    db = get_synthetic_db(input_directory)
    for name in db.getall():
        entry = db.get(name)
        matching_samples = []
        for sample in entry["samples"]:
            if sample["sample_size"] == sample_size and len(matching_samples) < runs:
                matching_samples.append(sample)
        if len(matching_samples) != runs:
            raise RuntimeError("Insufficient samples available, prepare more samples first")

        for sample in matching_samples:
            detail_learn_options = learn_options.copy()
            detail_learn_options.domain = os.path.join(input_directory, "{}.density".format(name))
            detail_learn_options.data = os.path.join(input_directory, sample["samples_file"])
            detail_learn_options.labels = os.path.join(input_directory, sample["labels_file"])

            export_file = "{}{sep}{}.{}.{}.result" \
                .format( output_directory, name, sample_size, sample["seed"], sep=os.path.sep)
            log_file = "{}{sep}{}.{}.{}.log" \
                .format(output_directory, name, sample_size, sample["seed"], sep=os.path.sep)

            if not os.path.exists(os.path.dirname(export_file)):
                os.makedirs(os.path.dirname(export_file))

            commands.append("incal-track {} --export {} --log {}"
                            .format(detail_learn_options.print_arguments(), export_file, log_file))

    run_commands(commands, processes, time_out)


def learn_benchmark(runs, sample_size, processes, time_out, learn_options: LearnOptions):
    # def filter1(entry):
    #     return "real_variables_count" in entry and entry["real_variables_count"] + entry["bool_variables_count"] <= 10
    #
    # count = 0
    # boolean = 0
    # for name, entry, density_filename in select_benchmark_files(filter1):
    #     if entry["bool_variables_count"] > 0:
    #         boolean += 1
    #     count += 1
    #
    # print("{} / {}".format(boolean, count))
    #
    # count = 0
    # boolean = 0
    # for name, entry, density_filename in select_benchmark_files(benchmark_filter):
    #     if entry["bool_variables_count"] > 0:
    #         boolean += 1
    #     count += 1
    #
    # print("{} / {}".format(boolean, count))

    def learn_filter(_e):
        return benchmark_filter(_e) and "samples" in _e

    count = 0
    problems_to_learn = []
    for name, entry, density_filename in select_benchmark_files(learn_filter):
        if len(entry["bounds"]) > 0:
            best_ratio = min(rel_ratio(t[1]) for t in entry["bounds"])
            if best_ratio <= 0.3:
                qualifying = [t for t in entry["bounds"] if rel_ratio(t[1]) <= 0.3 and abs(rel_ratio(t[1]) - best_ratio) <= best_ratio / 5]
                selected = sorted(qualifying, key=lambda x: get_bound_volume(x[0]))[0]
                print(name, "\n", rel_ratio(selected[1]), best_ratio, selected[0], entry["bool_variables_count"])
                count += 1
                selected_samples = [s for s in entry["samples"]
                                    if s["bounds"] == selected[0] and s["sample_size"] >= sample_size]
                if len(selected_samples) < runs:
                    raise RuntimeError("Insufficient number of data set available ({} of {})"
                                       .format(len(selected_samples), runs))
                elif len(selected_samples) > runs:
                    selected_samples = selected_samples[:runs]
                for selected_sample in selected_samples:
                    problems_to_learn.append((name, density_filename, selected_sample))

    commands = []
    for name, density_filename, selected_sample in problems_to_learn:
        detail_learn_options = learn_options.copy()
        detail_learn_options.domain = density_filename
        detail_learn_options.data = selected_sample["samples_filename"]
        detail_learn_options.labels = selected_sample["labels_filename"]
        export_file = "{}{sep}{}.{}.{}.result".format(get_benchmark_results_dir(), name, selected_sample["sample_size"],
                                                      selected_sample["seed"], sep=os.path.sep)
        log_file = "{}{sep}{}.{}.{}.log".format(get_benchmark_results_dir(), name, selected_sample["sample_size"],
                                                      selected_sample["seed"], sep=os.path.sep)
        if not os.path.exists(os.path.dirname(export_file)):
            os.makedirs(os.path.dirname(export_file))
        commands.append("incal-track {} --export {} --log {}"
                        .format(detail_learn_options.print_arguments(), export_file, log_file))

    run_commands(commands, processes, time_out)


def get_experiment(res_path=None):
    def import_handler(parameters_dict, results_dict, config_dict):
        for key, entry in parameters_dict.items():
            if isinstance(entry, str):
                index = entry.find("res/")
                if index >= 0:
                    parameters_dict[key] = res_path + os.path.sep + entry[index+4:]

    config = Options()
    config.add_option("export", str)
    return Experiment(LearnOptions(), LearnResults(), config, import_handler if res_path else None)


def track():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        experiment = get_experiment()
        experiment.import_from_command_line()
        experiment.save(experiment.config.export)
        experiment.execute()
        experiment.save(experiment.config.export)



from __future__ import print_function

from gurobipy.gurobipy import *


# noinspection PyPep8Naming
from learner import Learner


class KDNFLearner(Learner):
    def __init__(self, k, max_hyperplanes, alpha, negated_hyperplanes=True, min_covered=None, max_terms=None):
        self.k = k
        self.max_hyperplanes = max_hyperplanes
        self.alpha = alpha
        self.min_covered = min_covered
        self.max_terms = max_terms
        self.mu = 0.0005
        self.negated_hyperplanes = negated_hyperplanes

    def learn(self, domain, data):
        # Create a new model
        m = Model("kDNF")

        print("Data")
        for row, l in data:
            print(*(["{}: {:.2f}".format(v, Learner._convert(row[v])) for v in domain.variables] + [l]))
        print()

        # Computed constants
        n_f = len(domain.variables)
        n_h = (2 * self.max_hyperplanes) if self.negated_hyperplanes else self.max_hyperplanes
        n_d = len(data)
        k = self.k
        naive_misclassification = float(Learner._get_misclassification(data))

        x = [[Learner._convert(row[v]) for v in domain.variables] for row, _ in data]


        # Variables
        s_h = [m.addVar(vtype=GRB.BINARY, name="s_H({h})".format(h=h)) for h in range(n_h)]
        s_d = [m.addVar(vtype=GRB.BINARY, name="s_D({i})".format(i=i)) for i in range(n_d)]

        z_h = [
            [m.addVar(vtype=GRB.BINARY, name="z_H({h}, {c})".format(h=h, c=c)) for c in range(k)]
            for h in range(n_h)
        ]
        z_d = [
            [m.addVar(vtype=GRB.BINARY, name="z_D({i}, {c})".format(i=i, c=c)) for c in range(k)]
            for i in range(n_d)
        ]
        z_ih = [
            [m.addVar(vtype=GRB.BINARY, name="z_IH({i}, {h})".format(i=i, h=h)) for h in range(n_h)]
            for i in range(n_d)
        ]
        z_ihc = [
            [
                [m.addVar(vtype=GRB.BINARY, name="z_IHC({i}, {h}, {c})".format(i=i, h=h, c=c)) for c in range(k)]
                for h in range(n_h)
            ]
            for i in range(n_d)
        ]

        a = [
            [m.addVar(vtype=GRB.CONTINUOUS, lb=-1, ub=1, name="a({h}, {j})".format(h=h, j=j)) for j in range(n_f)]
            for h in range(n_h / 2 if self.negated_hyperplanes else n_h)
        ]
        b = [m.addVar(vtype=GRB.CONTINUOUS, lb=-1, ub=1, name="b({h})".format(h=h))
             for h in range(n_h / 2 if self.negated_hyperplanes else n_h)]
        print()

        # Set objective
        misclassification = sum(1 - s_d[i] if data[i][1] else s_d[i] for i in range(n_d))
        m.setObjective(misclassification / naive_misclassification + self.alpha * sum(s_h), GRB.MINIMIZE)

        # Add constraint: SUM_(j = 1..n_h) a(h, j) * x(i, j) <= b(h) + M(1 - z_ih(i, h)) for all h, i
        for h in range(n_h / 2 if self.negated_hyperplanes else n_h):
            for i in range(n_d):
                name = "SUM_(j = 1..n_H) a({h}, j) * x({i}, j) <= b({h}) + M(1 - z_IH({i}, {h}))".format(h=h, i=i)
                m.addConstr(sum(a[h][j] * x[i][j] for j in range(n_f)) <= b[h] + (n_f + 1) * (1 - z_ih[i][h]), name)
                print(name)
        print()

        for h in range(n_h / 2 if self.negated_hyperplanes else n_h):
            for i in range(n_d):
                name = "SUM_(j = 1..n_H) a({h}, j) * x({i}, j) >= b({h}) - mu - M * z_IH({i}, {h})".format(h=h, i=i)
                m.addConstr(sum(a[h][j] * x[i][j] for j in range(n_f)) >= b[h] - self.mu - (n_f + 1) * z_ih[i][h], name)
                print(name)
        print()

        if self.negated_hyperplanes:
            for h in range(n_h / 2):
                for i in range(n_d):
                    name = "z_IH({i}, {h}) = 1 - z_IH({i}, {hi})".format(i=i, h=h, hi=(h + n_h / 2))
                    m.addConstr(z_ih[i][n_h / 2 + h] == 1 - z_ih[i][h], name)
                    print(name)
            print()

        # If a hyperplane is assigned to any conjunction, it is selected (used)
        for h in range(n_h):
            for c in range(k):
                name = "s_H({h}) >= z_H({h}, {c})".format(h=h, c=c)
                m.addConstr(s_h[h] >= z_h[h][c], name)
                print(name)
        print()

        # If a hyperplane is not assigned to any conjunction, it cannot be selected (used)
        for h in range(n_h):
            name = "s_H({h}) <= SUM_(c = 1..k) z_H({h}, c)".format(h=h)
            m.addConstr(s_h[h] <= sum(z_h[h][c] for c in range(k)), name)
            print(name)
        print()

        # If an example i is not assigned to c, then it cannot be assigned to every h that is assigned to c
        # for c in range(k):
        #     for i in range(n_d):
        #         # TODO name
        #         name = "z_IH({i}, {h}) + z_H({h}, {c}) >= 2 - 2 * (1 - z_D({i}, {c}))".format(h=h, c=c, i=i)
        #         m.addConstr(sum(z_h[h][c] for h in range(n_h)) - sum(z_ihc[i][h][c] for h in range(n_h)) <= n_h * (1 - z_d[i][c]))
        #         m.addConstr(sum(z_h[h][c] for h in range(n_h)) - sum(z_ihc[i][h][c] for h in range(n_h)) >= n_h * (1 - z_d[i][c]))
        #         print(name)
        # print()

        # If (h, c) then (i, c) can only be selected if every (i, h, c) is selected
        for c in range(k):
            for h in range(n_h):
                for i in range(n_d):
                    name = "z_D({i}, {c}) <= z_IHC({i}, {h}, {c}) + (1 - z_H({h}, {c}))".format(c=c, h=h, i=i)
                    m.addConstr(z_d[i][c] <= z_ihc[i][h][c] + (1 - z_h[h][c]), name)
                    print(name)
        print()

        # If (h, c) then (i, c) can only not be selected if some (i, h, c) is not selected
        # for c in range(k):
        #     for h in range(n_h):
        #         for i in range(n_d):
        #             name = "z_D({i}, {c}) <= z_IHC({i}, {h}, {c}) + (1 - z_H({h}, {c}))".format(c=c, h=h, i=i)
        #             m.addConstr(z_d[i][c] <= z_ihc[i][h][c] + (1 - z_h[h][c]), name)
        #             print(name)
        # print()

        # If (i, c) is not selected, then there must exist an h such that: (h, c) and not (i, h, c)
        for c in range(k):
            for i in range(n_d):
                name = "..."  # TODO name
                m.addConstr(sum(z_h[h][c] for h in range(n_h)) - sum(z_ihc[i][h][c] for h in range(n_h)) >= 1 - z_d[i][c], name)
                print(name)
        print()

        # If (i, c) is selected, then there must be at least one selected tuple (h, c)
        for c in range(k):
            for i in range(n_d):
                name = "SUM_(h = 1..n_H) z_H(h, {c}) >= 1 - (1 - z_D({i}, {c}))".format(c=c, i=i)
                m.addConstr(sum(z_h[h][c] for h in range(n_h)) >= 1 - (1 - z_d[i][c]))
                print(name)
        print()

        # If the triple (i, h, c) is selected, then both (i, h) and (h, c) must be true
        for h in range(n_h):
            for c in range(k):
                for i in range(n_d):
                    name = "z_IH({i}, {h}) + z_H({h}, {c}) >= 2 - 2 * (1 - z_IHC({i}, {h}, {c}))".format(h=h, c=c, i=i)
                    m.addConstr(z_ih[i][h] + z_h[h][c] >= 2 - 2 * (1 - z_ihc[i][h][c]), name)
                    print(name)
        print()

        # If the triple (i, h, c) is not selected, then either (i, h) or (h, c) must be false
        for h in range(n_h):
            for c in range(k):
                for i in range(n_d):
                    name = "z_IH({i}, {h}) + z_H({h}, {c}) <= 2 * z_IHC({i}, {h}, {c}) + 1".format(h=h, c=c, i=i)
                    m.addConstr(z_ih[i][h] + z_h[h][c] <= 2 * z_ihc[i][h][c] + 1, name)
                    print(name)
        print()

        # If an example is assigned to any conjunction, it is selected (covered)
        for i in range(n_d):
            for c in range(k):
                name = "s_D({i}) >= z_D({i}, {c})".format(i=i, c=c)
                m.addConstr(s_d[i] >= z_d[i][c], name)
                print(name)
        print()

        # If an example is not assigned to any conjunction, it cannot be selected (covered)
        for i in range(n_d):
            name = "s_D({i}) <= SUM_(c = 1..k) z_D({i}, c)".format(i=i)
            m.addConstr(s_d[i] <= sum(z_d[i][c] for c in range(k)), name)
            print(name)
        print()

        # Examples can only be assigned to active hyperplanes
        # z_ih <= s_h for all i, h
        # for h in range(n_h):
        #     for i in range(n_d):
        #         name = "z_ih({i}, {h}) <= s_h({h})".format(i=i, h=h)
        #         m.addConstr(z_ih[i][h] <= s_h[h], name)
        #         print(name)
        # print()

        # Hack example
        # m.addConstr(a[0][0] == 1)
        # m.addConstr(a[0][1] == 0)
        # m.addConstr(b[0] == 0.5)
        #
        # m.addConstr(a[1][0] == 0)
        # m.addConstr(a[1][1] == 1)
        # m.addConstr(b[1] == 0.5)
        #
        # m.addConstr(a[2][0] == -1)
        # m.addConstr(a[2][1] == 0)
        # m.addConstr(b[2] == -0.5)
        #
        # m.addConstr(a[3][0] == 0)
        # m.addConstr(a[3][1] == -1)
        # m.addConstr(b[3] == -0.5)
        #
        # m.addConstr(s_h[0] == 1)
        # m.addConstr(s_h[1] == 1)
        # m.addConstr(s_h[2] == 1)
        # m.addConstr(s_h[3] == 1)

        # m.addConstr(z_h[0][0] == 1)
        # m.addConstr(z_h[1][0] == 1)
        # m.addConstr(z_h[2][0] == 0)
        # m.addConstr(z_h[3][0] == 0)
        # m.addConstr(z_h[0][1] == 0)
        # m.addConstr(z_h[1][1] == 0)
        # m.addConstr(z_h[2][1] == 1)
        # m.addConstr(z_h[3][1] == 1)

        # m.addConstr(z_ih[0][0] == 1)
        # m.addConstr(z_ih[1][0] == 0)
        # m.addConstr(z_ih[2][0] == 1)
        # m.addConstr(z_ih[3][0] == 0)
        # m.addConstr(z_ih[0][1] == 1)
        # m.addConstr(z_ih[1][1] == 0)
        # m.addConstr(z_ih[2][1] == 0)
        # m.addConstr(z_ih[3][1] == 1)
        # m.addConstr(z_ih[0][2] == 0)
        # m.addConstr(z_ih[1][2] == 1)
        # m.addConstr(z_ih[2][2] == 0)
        # m.addConstr(z_ih[3][2] == 1)
        # m.addConstr(z_ih[0][3] == 0)
        # m.addConstr(z_ih[1][3] == 1)
        # m.addConstr(z_ih[2][3] == 1)
        # m.addConstr(z_ih[3][3] == 0)

        # m.addConstr(z_d[0][0] == 1)
        # m.addConstr(z_d[1][0] == 0)
        # m.addConstr(z_d[2][0] == 0)
        # m.addConstr(z_d[3][0] == 0)
        # m.addConstr(z_d[0][1] == 0)
        # m.addConstr(z_d[1][1] == 0)
        # m.addConstr(z_d[2][1] == 1)
        # m.addConstr(z_d[3][1] == 0)

        # Wrong solution
        # m.addConstr(a[0][0] == 1)
        # m.addConstr(a[0][1] == 0)
        # m.addConstr(b[0] == 0.5)
#
        # m.addConstr(a[1][0] == 0)
        # m.addConstr(a[1][1] == 1)
        # m.addConstr(b[1] == 0.5)
#
        # m.addConstr(a[2][0] == -1)
        # m.addConstr(a[2][1] == 0)
        # m.addConstr(b[2] == -0.6)
#
        # m.addConstr(a[3][0] == 0)
        # m.addConstr(a[3][1] == -1)
        # m.addConstr(b[3] == -0.6)
#
        # m.addConstr(s_h[0] == 1)
        # m.addConstr(s_h[1] == 1)
        # m.addConstr(s_h[2] == 1)
        # m.addConstr(s_h[3] == 1)


        m.optimize()

        for v in m.getVars():
            print(v.varName, v.x)

        print('Obj:', m.objVal)

        print()

        # Lets extract the kDNF formulation
        print("Hyperplanes")
        conjunctions = []
        from pysmt.shortcuts import And, Or, Real, LE, Plus, Times, Symbol

        for conj in range(k):
            conjunction = []
            for h in range(n_h):
                if int(z_h[h][conj].x) == 1:
                    if h < n_h / 2 or not self.negated_hyperplanes:
                        coefficients = [Real(float(a[h][j].x)) for j in range(n_f)]
                        constant = Real(float(b[h].x))
                    else:
                        coefficients = [Real(-float(a[h - n_h / 2][j].x)) for j in range(n_f)]
                        constant = Real(-float(b[h - n_h / 2].x))
                    linear_sum = Plus([Times(c, domain.get_symbol(v)) for c, v in zip(coefficients, domain.variables)])
                    conjunction.append(LE(linear_sum, constant))
            conjunctions.append(conjunction)
        return conjunctions


    @staticmethod
    def _get_misclassification(data):
        true_count = 0
        for _, l in data:
            if l:
                true_count += 1
        return min(true_count, len(data) - true_count)

    @staticmethod
    def _convert(value):
        return float(value.constant_value())



import os
import signal
import subprocess
from multiprocessing.pool import Pool
from subprocess import TimeoutExpired


def run_command(args):
    command, time_out = args
    with subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, preexec_fn=os.setsid) as process:
        try:
            process.communicate(timeout=time_out)
        except TimeoutExpired:
            os.killpg(process.pid, signal.SIGINT)  # send signal to the process group
            process.communicate()


def run_commands(commands, processes=None, time_out=None):
    pool = Pool(processes=processes)
    commands = [(command, time_out) for command in commands]
    pool.map(run_command, commands)



from __future__ import print_function, division

import argparse
import hashlib
import json
import random

import os
import tempfile

import time

import problem
import generator
import parse
import inc_logging

from os.path import basename

import pysmt.shortcuts as smt

from incremental_learner import RandomViolationsStrategy
from k_cnf_smt_learner import KCnfSmtLearner
from parameter_free_learner import learn_bottom_up


def learn(name, domain, h, data, seed):
    initial_size = 20
    violations_size = 10
    log_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "demo", "results")
    problem_name = hashlib.sha256(name).hexdigest()

    def learn_inc(_data, _k, _h):
        violations_strategy = RandomViolationsStrategy(violations_size)
        learner = KCnfSmtLearner(_k, _h, violations_strategy)
        initial_indices = random.sample(list(range(len(data))), initial_size)
        log_file = os.path.join(log_dir, "{}_{}_{}.txt".format(problem_name, _k, _h))
        learner.add_observer(inc_logging.LoggingObserver(log_file, seed, True, violations_strategy))
        learned_theory = learner.learn(domain, data, initial_indices)
        # learned_theory = Or(*[And(*planes) for planes in hyperplane_dnf])
        print("Learned theory:\n{}".format(parse.smt_to_nested(learned_theory)))
        return learned_theory

    phi, k, h = learn_bottom_up(data, learn_inc, 1, 1, init_h=h, max_h=h)

    with open(os.path.join(log_dir, "problems.txt"), "a") as f:
        print(json.dumps({problem_name: name, "k": k, "h": h}), file=f)


def main(filename, sample_count):
    seed = time.time()
    random.seed(seed)

    target_formula = smt.read_smtlib(filename)

    variables = target_formula.get_free_variables()
    var_names = [str(v) for v in variables]
    var_types = {str(v): v.symbol_type() for v in variables}
    var_domains = {str(v): (0, 200) for v in variables}  # TODO This is a hack

    domain = problem.Domain(var_names, var_types, var_domains)
    name = basename(filename).split(".")[0]
    target_problem = problem.Problem(domain, target_formula, name)

    # compute_difference(domain, target_formula, target_formula)

    samples = generator.get_problem_samples(target_problem, sample_count, 1)

    initial_indices = random.sample(list(range(sample_count)), 20)
    learner = KCnfSmtLearner(3, 3, RandomViolationsStrategy(5))

    dir_name = "../output/{}".format(name)
    img_name = "{}_{}_{}".format(learner.name, sample_count, seed)
    # learner.add_observer(plotting.PlottingObserver(data_set.samples, dir_name, img_name, "r0", "r1"))
    with open("log.txt", "w") as f:
        learner.add_observer(inc_logging.LoggingObserver(f))

        print(parse.smt_to_nested(learner.learn(domain, samples, initial_indices)))


def compute_difference(domain, target_theory, learned_theory):
    query = (target_theory & ~learned_theory) | (~target_theory & learned_theory)
    compute_wmi(domain, query, domain.variables)


def compute_wmi(domain, query, variables):
    # os.environ["PATH"] += os.pathsep + "/Users/samuelkolb/Downloads/latte/dest/bin"
    # from sys import path
    # path.insert(0, "/Users/samuelkolb/Documents/PhD/wmi-pa/src")
    # from wmi import WMI

    # support = []
    # for v in domain.real_vars:
    #     lb, ub = domain.var_domains[v]
    #     sym = domain.get_symbol(v)
    #     support.append((lb <= sym) & (sym <= ub))
    #
    # support = smt.And(*support)
    # wmi = WMI()
    # total_volume, _ = wmi.compute(support, 1, WMI.MODE_PA)
    # query_volume, _ = wmi.compute(support & query, 1, WMI.MODE_PA)
    # print(query_volume / total_volume)

    f = tempfile.NamedTemporaryFile(delete=False)
    try:
        flat = {
            "domain": problem.export_domain(domain, to_str=False),
            "query": parse.smt_to_nested(query),
            "variables": variables
        }
        json.dump(flat, f)
        with open("test.txt", "w") as f2:
            json.dump(flat, f2)
        f.close()
    finally:
        os.remove(f.name)



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("filename")
    parser.add_argument("sample_count", type=int)
    args = parser.parse_args()
    main(args.filename, args.sample_count)



import random

import numpy as np
from incal.observe.inc_logging import LoggingObserver
from pysmt.fnode import FNode
from pywmi import smt_to_nested
from pywmi.domain import Density, Domain
from typing import Tuple, Optional

from .parameter_free_learner import learn_bottom_up
from .violations.core import RandomViolationsStrategy
from .violations.dt_selection import DecisionTreeSelection
from .k_cnf_smt_learner import KCnfSmtLearner
from .util.options import Options, Results


class LearnOptions(Options):
    def __init__(self):
        super().__init__(learn)
        self.add_option("domain", str, None, LearnOptions.domain_extraction)
        self.add_option("data", str, None, LearnOptions.np_extraction)
        self.add_option("labels", str, None, LearnOptions.np_extraction)

        self.add_option("learner", (str, str), ("cnf", "-"), Options.convert_dict(
            cnf=LearnOptions.cnf_factory_wrap
        ), arg_name="learner_factory")
        self.add_option("initial_strategy", (str, int), ("random", 20), Options.convert_dict(
            random=LearnOptions.initial_random
        ))
        self.add_option("selection_strategy", (str, int), ("random", 10), Options.convert_dict(
            random=LearnOptions.select_random,
            dt=LearnOptions.select_dt
        ))
        self.add_option("initial_k", int, 1)
        self.add_option("initial_h", int, 0)
        self.add_option("weight_k", float, 1)
        self.add_option("weight_h", float, 1)
        self.add_option("log", str)
        # self.add_option("max_k", int, None)
        # self.add_option("max_h", int, None)

    @staticmethod
    def domain_extraction(filename):
        return Density.from_file(filename).domain

    @staticmethod
    def np_extraction(filename):
        return np.load(filename)

    @staticmethod
    def cnf_factory_wrap(symmetries):
        def cnf_factory(k, h, selection_strategy):
            return KCnfSmtLearner(k, h, selection_strategy, symmetries=symmetries)
        return cnf_factory

    @staticmethod
    def initial_random(count):
        def random_selection(indices):
            return random.sample(indices, count)
        return random_selection

    @staticmethod
    def select_random(count):
        return RandomViolationsStrategy(count)

    @staticmethod
    def select_dt(_count):
        return DecisionTreeSelection()

    def make_copy(self):
        return LearnOptions()


class LearnResults(Results):
    def __init__(self):
        super().__init__()
        self.add_duration()
        self.add_result("formula", LearnResults.extract_formula)
        self.add_result("k", LearnResults.extract_k)
        self.add_result("h", LearnResults.extract_h)

    @staticmethod
    def extract_formula(result):
        return smt_to_nested(result[0])

    @staticmethod
    def extract_k(result):
        return result[1]

    @staticmethod
    def extract_h(result):
        return result[2]


def learn(
        domain: Domain,
        data: np.ndarray,
        labels: np.ndarray,
        learner_factory: callable,
        initial_strategy: callable,
        selection_strategy: object,
        initial_k: int,
        initial_h: int,
        weight_k: float,
        weight_h: float,
        log: Optional[str]=None
) -> Tuple[FNode, int, int]:
    """
    Learn a formula that separates the positive and negative examples
    :return: A tuple containing 1. the learned formula, 2. the number of terms (or clauses) used,
    3. the number of hyperplanes used
    """

    # log_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "demo", "results")
    # problem_name = hashlib.sha256(name).hexdigest()

    def learn_inc(_data, _labels, _i, _k, _h):
        learner = learner_factory(_k, _h, selection_strategy)
        initial_indices = initial_strategy(list(range(len(_data))))
        # log_file = os.path.join(log_dir, "{}_{}_{}.txt".format(problem_name, _k, _h))
        if log is not None:
            learner.add_observer(LoggingObserver(log, _k, _h, None, False, selection_strategy))
        return learner.learn(domain, _data, _labels, initial_indices)

    ((_d, _l, formula), k, h) =\
        learn_bottom_up(data, labels, learn_inc, weight_k, weight_h, initial_k, initial_h, None, None)
    return formula, k, h



from __future__ import print_function

from learner import Learner
import gurobipy.gurobipy as milp


class GreedyRuleLearner(Learner):
    def __init__(self, rule_learner_name, max_hyperplanes_per_rule, max_terms_per_rule):
        Learner.__init__(self, "greedy_{}".format(rule_learner_name))
        self.max_hyperplanes_per_rule = max_hyperplanes_per_rule
        self.max_terms_per_rule = max_terms_per_rule

    def learn(self, domain, data, border_indices):
        pos_uncovered_indices = []
        neg_indices = []
        for i in range(len(data)):
            if data[i][1]:
                pos_uncovered_indices.append(i)
            else:
                neg_indices.append(i)

        dnf_list = []
        iterations = 0
        while len(pos_uncovered_indices) > 0:
            print("{} uncovered positive examples remain".format(len(pos_uncovered_indices)))
            rule_list = self.learn_rule(domain, data, pos_uncovered_indices, neg_indices)
            new_pos_uncovered_indices = []
            for i in pos_uncovered_indices:
                if not Learner.check_example(domain, data[i][0], [rule_list]):
                    new_pos_uncovered_indices.append(i)
            dnf_list.append(rule_list)
            print("Covered {} new positive examples".format(len(pos_uncovered_indices) - len(new_pos_uncovered_indices)))
            pos_uncovered_indices = new_pos_uncovered_indices
            print(rule_list)
            print(pos_uncovered_indices)
            iterations += 1
            if iterations > 0:
                break
        print("All positive examples covered")

        for i in neg_indices:
            if Learner.check_example(domain, data[i][0], dnf_list):
                print("Negative example {}: {} was wrongfully covered".format(i, data[i][0]))
        return dnf_list

    def learn_rule(self, domain, data, pos_uncovered_indices, neg_indices):
        raise NotImplementedError()


class GreedyMilpRuleLearner(GreedyRuleLearner):
    def __init__(self, max_hyperplanes_per_rule, max_terms_per_rule):
        GreedyRuleLearner.__init__(self, "milp", max_hyperplanes_per_rule, max_terms_per_rule)

    def learn_rule(self, domain, data, pos_uncovered_indices, neg_indices):
        # Create a new model
        m = milp.Model("kDNF")

        print("Data")
        for row, l in data:
            print(*(["{}: {:.2f}".format(v, Learner._convert(row[v])) for v in domain.variables] + [l]))
        print()

        # --- Computed constants
        n_r = len(domain.real_vars)
        n_b = len(domain.bool_vars)
        n_h = self.max_hyperplanes_per_rule
        n_e = len(data)
        k = self.max_terms_per_rule
        mu = 0.005

        x_er = [[Learner._convert(row[v]) for v in domain.real_vars] for row, _ in data]
        x_eb = [[Learner._convert(row[v]) for v in domain.bool_vars] for row, _ in data]

        # --- Variables

        # Inequality h is selected for the rule
        s_h = [m.addVar(vtype=milp.GRB.BINARY, name="s_h({h})".format(h=h)) for h in range(n_h)]

        # Example e covered by inequality h
        c_eh = [
            [m.addVar(vtype=milp.GRB.BINARY, name="c_eh({e}, {h})".format(e=e, h=h)) for h in range(n_h)]
            for e in range(n_e)
        ]

        # Example e covered by rule
        c_e = [m.addVar(vtype=milp.GRB.BINARY, name="c_e({e})".format(e=e)) for e in range(n_e)]

        # Boolean feature (or negation) selected
        s_b = [m.addVar(vtype=milp.GRB.BINARY, name="s_b({b})".format(b=b)) for b in range(2 * n_b)]

        # Coefficient of the r-th variable of inequality h
        a_hr = [
            [m.addVar(vtype=milp.GRB.CONTINUOUS, lb=-1, ub=1, name="a_hr({h}, {r})".format(h=h, r=r))
             for r in range(n_r)]
            for h in range(n_h)
        ]

        # Offset of inequality h
        b_h = [m.addVar(vtype=milp.GRB.CONTINUOUS, lb=-1, ub=1, name="b_h({h})".format(h=h)) for h in range(n_h)]
        b_abs_h = [m.addVar(vtype=milp.GRB.CONTINUOUS, lb=-1, ub=1, name="b_abs_h({h})".format(h=h)) for h in range(n_h)]

        # Auxiliary variable (c_eh AND s_h)
        and_eh = [
            [m.addVar(vtype=milp.GRB.BINARY, name="and_eh({e}, {h})".format(e=e, h=h)) for h in range(n_h)]
            for e in range(n_e)
        ]

        # Auxiliary variable (x_eb AND s_b)
        and_eb = [
            [m.addVar(vtype=milp.GRB.BINARY, name="and_eb({e}, {b})".format(e=e, b=b)) for b in range(n_b)]
            for e in range(n_e)
        ]

        print()

        # Set objective
        m.setObjective(sum(c_e[e] for e in range(n_e) if data[e][1]), milp.GRB.MAXIMIZE)

        # Add constraint: and_eh <= c_eh
        for e in range(n_e):
            if not data[e][1]:
                for h in range(n_h):
                    name = "and_eh({e}, {h}) <= c_eh({e}, {h})".format(e=e, h=h)
                    m.addConstr(and_eh[e][h] <= c_eh[e][h], name)
                    print(name)
        print()

        # Add constraint: and_eh <= s_h
        for e in range(n_e):
            if not data[e][1]:
                for h in range(n_h):
                    name = "and_eh({e}, {h}) <= s_h({h})".format(e=e, h=h)
                    m.addConstr(and_eh[e][h] <= s_h[h], name)
                    print(name)
        print()

        # Add constraint: and_eh >= c_eh + s_h - 1
        for e in range(n_e):
            if not data[e][1]:
                for h in range(n_h):
                    name = "and_eh({e}, {h}) >= c_eh({e}, {h}) + s_h({h}) - 1".format(e=e, h=h)
                    m.addConstr(and_eh[e][h] >= c_eh[e][h] + s_h[h] - 1, name)
                    print(name)
        print()

        # Add constraint: and_eb <= x_eb
        for e in range(n_e):
            if not data[e][1]:
                for b in range(2 * n_b):
                    name = "and_eb({e}, {b}) <= x_eb({e}, {b})".format(e=e, b=b)
                    m.addConstr(and_eb[e][b] <= (x_eb[e][b] if b < n_b else 1 - x_eb[e][b]), name)
                    print(name)
        print()

        # Add constraint: and_eb <= s_h
        for e in range(n_e):
            if not data[e][1]:
                for b in range(2 * n_b):
                    name = "and_eb({e}, {b}) <= s_b({b})".format(e=e, b=b)
                    m.addConstr(and_eb[e][b] <= s_b[b], name)
                    print(name)
        print()

        # Add constraint: and_eb >= x_eb + s_b - 1
        for e in range(n_e):
            if not data[e][1]:
                for b in range(2 * n_b):
                    name = "and_eb({e}, {b}) >= x_eb({e}, {b}) + s_b({b}) - 1".format(e=e, b=b)
                    m.addConstr(and_eb[e][b] >= (x_eb[e][b] if b < n_b else 1 - x_eb[e][b]) + s_b[b] - 1, name)
                    print(name)
        print()

        # Add constraint: SUM_(h = 1..n_h) and_eh - s_h + SUM_(b = 1..2 * n_b) and_eb - s_b <= -1
        for e in range(n_e):
            if not data[e][1]:
                name = "SUM_(h = 1..n_h) and_eh({e}, h) - s_h(h) + SUM_(b = 1..2*n_b) and_eb({e}, b) - s_b(b) <= -1"\
                    .format(e=e)
                m.addConstr(sum(and_eh[e][h] - s_h[h] for h in range(n_h))
                            + sum(and_eb[e][b] - s_b[b] for b in range(2 * n_b)) <= -1, name)
                print(name)

        # Add constraint: c_e <= c_eh + (1 - s_h)
        for e in range(n_e):
            if data[e][1]:
                for h in range(n_h):
                    name = "c_e({e}) <= c_eh({e}, {h}) + (1 - s_h({h}))".format(e=e, h=h)
                    m.addConstr(c_e[e] <= c_eh[e][h] + (1 - s_h[h]), name)
                    print(name)

        # Add constraint: c_e <= x_eb + (1 - s_b)
        for e in range(n_e):
            if data[e][1]:
                for b in range(2 * n_b):
                    name = "c_e({e}) <= x_eb({e}, {b}) + (1 - s_b({b}))".format(e=e, b=b)
                    m.addConstr(c_e[e] <= x_eb[e][b] + (1 - s_b[b]), name)
                    print(name)

        # Add constraint: SUM_(r = 1..n_r) a_hr * x_er <= b_h + 2 * (1 - c_eh)
        for e in range(n_e):
            for h in range(n_h):
                name = "SUM_(r = 1..n_r) a_hr({h}, r) * x_er({e}, r) <= b_h({h}) + 2 * (1 - c_eh({e}, {h}))"\
                    .format(e=e, h=h)
                m.addConstr(sum(a_hr[h][r] * x_er[e][r] for r in range(n_r)) <= b_h[h] + 2 * (1 - c_eh[e][h]), name)
                print(name)

        # Add constraint: SUM_(r = 1..n_r) a_hr * x_er >= b_h - mu - 2 * c_eh
        for e in range(n_e):
            for h in range(n_h):
                name = "SUM_(r = 1..n_r) a_hr({h}, r) * x_er({e}, r) >= b_h({h}) - mu - 2 * c_eh({e}, {h})"\
                    .format(e=e, h=h)
                m.addConstr(sum(a_hr[h][r] * x_er[e][r] for r in range(n_r)) >= b_h[h] + mu - 2 * c_eh[e][h], name)
                print(name)

        # Add constraint: SUM_(h = 1..n_h) s_h + SUM_(b = 1..n_b) s_b <= k
        name = "SUM_(h = 1..n_h) s_h(h) + SUM_(b = 1..n_b) s_b(b) <= k"
        m.addConstr(sum(s_h[h] for h in range(n_h)) + sum(s_b[b] for b in range(2 * n_b)) <= k, name)
        print(name)

        # Hack example
        # m.addConstr(a[0][0] == 1)
        # m.addConstr(a[0][1] == 0)
        # m.addConstr(b[0] == 0.5)
        #
        # m.addConstr(a[1][0] == 0)
        # m.addConstr(a[1][1] == 1)
        # m.addConstr(b[1] == 0.5)
        #
        # m.addConstr(a[2][0] == -1)
        # m.addConstr(a[2][1] == 0)
        # m.addConstr(b[2] == -0.5)
        #
        # m.addConstr(a[3][0] == 0)
        # m.addConstr(a[3][1] == -1)
        # m.addConstr(b[3] == -0.5)
        #
        # m.addConstr(s_h[0] == 1)
        # m.addConstr(s_h[1] == 1)
        # m.addConstr(s_h[2] == 1)
        # m.addConstr(s_h[3] == 1)

        # m.addConstr(z_h[0][0] == 1)
        # m.addConstr(z_h[1][0] == 1)
        # m.addConstr(z_h[2][0] == 0)
        # m.addConstr(z_h[3][0] == 0)
        # m.addConstr(z_h[0][1] == 0)
        # m.addConstr(z_h[1][1] == 0)
        # m.addConstr(z_h[2][1] == 1)
        # m.addConstr(z_h[3][1] == 1)

        # m.addConstr(z_ih[0][0] == 1)
        # m.addConstr(z_ih[1][0] == 0)
        # m.addConstr(z_ih[2][0] == 1)
        # m.addConstr(z_ih[3][0] == 0)
        # m.addConstr(z_ih[0][1] == 1)
        # m.addConstr(z_ih[1][1] == 0)
        # m.addConstr(z_ih[2][1] == 0)
        # m.addConstr(z_ih[3][1] == 1)
        # m.addConstr(z_ih[0][2] == 0)
        # m.addConstr(z_ih[1][2] == 1)
        # m.addConstr(z_ih[2][2] == 0)
        # m.addConstr(z_ih[3][2] == 1)
        # m.addConstr(z_ih[0][3] == 0)
        # m.addConstr(z_ih[1][3] == 1)
        # m.addConstr(z_ih[2][3] == 1)
        # m.addConstr(z_ih[3][3] == 0)

        # m.addConstr(z_d[0][0] == 1)
        # m.addConstr(z_d[1][0] == 0)
        # m.addConstr(z_d[2][0] == 0)
        # m.addConstr(z_d[3][0] == 0)
        # m.addConstr(z_d[0][1] == 0)
        # m.addConstr(z_d[1][1] == 0)
        # m.addConstr(z_d[2][1] == 1)
        # m.addConstr(z_d[3][1] == 0)

        # Wrong solution
        # m.addConstr(a[0][0] == 1)
        # m.addConstr(a[0][1] == 0)
        # m.addConstr(b[0] == 0.5)
        #
        # m.addConstr(a[1][0] == 0)
        # m.addConstr(a[1][1] == 1)
        # m.addConstr(b[1] == 0.5)
        #
        # m.addConstr(a[2][0] == -1)
        # m.addConstr(a[2][1] == 0)
        # m.addConstr(b[2] == -0.6)
        #
        # m.addConstr(a[3][0] == 0)
        # m.addConstr(a[3][1] == -1)
        # m.addConstr(b[3] == -0.6)
        #
        # m.addConstr(s_h[0] == 1)
        # m.addConstr(s_h[1] == 1)
        # m.addConstr(s_h[2] == 1)
        # m.addConstr(s_h[3] == 1)

        # m.addConstr(s_h[0] == 1)
        # m.addConstr(s_h[1] == 1)
        # m.addConstr(s_h[2] == 0)
        # m.addConstr(s_h[3] == 0)
#
        # m.addConstr(a_hr[0][0] == -1)
        # m.addConstr(a_hr[0][1] == -0.5)
        # m.addConstr(b_h[0] == -0.2)
#
        # m.addConstr(a_hr[1][0] == 1)
        # m.addConstr(a_hr[1][1] == 0)
        # m.addConstr(b_h[1] == 0.75)

        m.optimize()

        for v in m.getVars():
            print(v.varName, v.x)

        print('Obj:', m.objVal)

        print()

        # Lets extract the rule formulation
        rule = []
        from pysmt.shortcuts import Real, LE, Plus, Times, Not

        for h in range(n_h):
            if int(s_h[h].x) == 1:
                coefficients = [Real(float(a_hr[h][r].x)) for r in range(n_r)]
                constant = Real(float(b_h[h].x))
                linear_sum = Plus([Times(c, domain.get_symbol(v)) for c, v in zip(coefficients, domain.real_vars)])
                rule.append(LE(linear_sum, constant))

        for b in range(2 * n_b):
            if int(s_b[b].x) == 1:
                var = domain.get_symbol(domain.bool_vars[b if b < n_b else b - n_b])
                rule.append(var if b < n_b else Not(var))

        print("Features", x_er[4][0], x_er[4][1])
        print(sum(a_hr[1][r].x * x_er[4][r] for r in range(n_r)), "<=", b_h[1].x + 2 * (1 - c_eh[4][1].x))
        print(sum(a_hr[1][r].x * x_er[4][r] for r in range(n_r)), ">=", b_h[1].x + mu - 2 * c_eh[4][1].x)

        return rule




import argparse
import os
from glob import glob

from autodora.observe import ProgressObserver
from autodora.observers.telegram_observer import TelegramObserver
from autodora.runner import CommandLineRunner, PrintObserver
from autodora.sql_storage import SqliteStorage
from autodora.trajectory import product

try:
    from telegram.ext import Updater
    from telegram import ParseMode
except ImportError:
    Updater = None
    ParseMode = None

from .learn_experiment import LearnExperiment


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "dirs",
        nargs="+",
        type=str,
        help="The directories to load files from to run experiments on",
    )

    parser.add_argument("--learners", nargs="+", type=str, help="The learners to use")
    parser.add_argument(
        "--dry",
        action="store_true",
        help="Performs a dry run, only printing out experiments instead of running them",
    )
    parser.add_argument(
        "--processes",
        type=int,
        default=None,
        help="The number of concurrent processes to use for running experiments.",
    )

    args = parser.parse_args()

    file_settings = dict()
    log_files = dict()
    storage = SqliteStorage()
    processes = args.processes or 1

    dispatcher = ProgressObserver()
    dispatcher.add_observer(PrintObserver())
    try:
        dispatcher.add_observer(TelegramObserver())
    except RuntimeError:
        pass

    for directory in args.dirs:
        directory = os.path.abspath(directory)
        dir_files = []
        for filename in glob(os.path.join(directory, "*.json")):
            basename = ".".join(filename.split(".")[:-1])
            for data_filename in glob(f"{basename}*.data.npy"):
                label_filename = data_filename.replace("data", "labels")
                dir_files.append((filename, data_filename, label_filename))

        name = os.path.basename(directory)
        file_settings[name] = []
        log_files[name] = os.path.join(directory, "run_exp.log")
        for f, d, l in dir_files:
            file_settings[name].append(
                {"formula_filename": f, "data_filename": d, "labels_filename": l}
            )

    learners_settings = {"learner": args.learners or ["cnf."]}

    for key, values in file_settings.items():
        trajectory = LearnExperiment.explore(key, product(values, learners_settings))
        if args.dry:
            print(*trajectory.experiments, sep="\n")
        else:
            CommandLineRunner(
                trajectory,
                storage,
                processes=processes,
                timeout=100,
                observer=dispatcher,
            ).run()


if __name__ == "__main__":
    main()



from pywmi import SmtWalker, smt_to_nested


class HalfSpaceWalker(SmtWalker):
    def __init__(self):
        self.half_spaces = set()

    def walk_and(self, args):
        self.walk_smt_multiple(args)

    def walk_or(self, args):
        self.walk_smt_multiple(args)

    def walk_plus(self, args):
        self.walk_smt_multiple(args)

    def walk_minus(self, left, right):
        self.walk_smt_multiple([left, right])

    def walk_times(self, args):
        self.walk_smt_multiple(args)

    def walk_not(self, argument):
        self.walk_smt_multiple([argument])

    def walk_ite(self, if_arg, then_arg, else_arg):
        self.walk_smt_multiple([if_arg, then_arg, else_arg])

    def walk_pow(self, base, exponent):
        self.walk_smt_multiple([base, exponent])

    def walk_lte(self, left, right):
        self.half_spaces.add(smt_to_nested(left <= right))

    def walk_lt(self, left, right):
        self.half_spaces.add(smt_to_nested(left < right))

    def walk_equals(self, left, right):
        self.walk_smt_multiple([left, right])

    def walk_symbol(self, name, v_type):
        pass

    def walk_constant(self, value, v_type):
        pass

    def find_half_spaces(self, formula):
        self.walk_smt(formula)
        return list(self.half_spaces)



from pywmi.smt_print import pretty_print

from .learn import LearnOptions


def main():
    formula, k, h = LearnOptions().execute_from_command_line("Learn SMT(LRA) theories from data")
    print("Learned formula (k={k}, h={h}): {f}".format(f=pretty_print(formula), k=k, h=h))




import argparse
import glob
import hashlib
import itertools
import json
import os
import pickle
import random
import shutil
import time
import traceback
import urllib.request
import zipfile
from collections import defaultdict
from typing import Tuple

import numpy as np
import pickledb
import pysmt.shortcuts as smt
import pysmt.environment
from pywmi import Domain, smt_to_nested, evaluate, RejectionEngine, export_domain, nested_to_smt
from pywmi.domain import Density, import_domain
from pywmi.sample import uniform

from .find_hyperplanes import HalfSpaceWalker
from .find_operators import OperatorWalker


def get_res_root(*args):
    return os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))), "res", *args)


def get_benchmark_dir():
    return get_res_root("smt_lib_benchmark")


def get_benchmark_cache_dir():
    return os.path.join(get_benchmark_dir(), "qf_lra_cache")


def get_benchmark_samples_dir():
    return os.path.join(get_benchmark_dir(), "qf_lra_samples")


def get_benchmark_results_dir():
    return os.path.join(get_benchmark_dir(), "qf_lra_results")


def get_summary_file():
    return os.path.join(get_benchmark_dir(), "qf_lra_summary.pickle")


# https://stackoverflow.com/a/11385480/253387
def fix_zip_file(zip_file):
    with open(zip_file, 'r+b') as f:
        data = f.read()
        pos = data.find(b'\x50\x4b\x05\x06')  # End of central directory signature
        if pos > 0:
            f.seek(pos + 22)  # size of 'ZIP end of central directory record'
            f.truncate()
        else:
            pass


def checksum(filename):
    hash_engine = hashlib.sha512()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_engine.update(chunk)
    return hash_engine.hexdigest()


def import_problem(filename):
    target_formula = smt.read_smtlib(filename)

    variables = target_formula.get_free_variables()
    var_names = [str(v) for v in variables]
    var_types = {str(v): v.symbol_type() for v in variables}
    var_domains = {str(v): (None, None) for v in variables}  # TODO This is a hack

    return (Domain(var_names, var_types, var_domains)), target_formula


def prepare_smt_lib_benchmark():
    benchmark_folder = get_res_root("smt_lib_benchmark")
    if not os.path.exists(benchmark_folder):
        os.makedirs(benchmark_folder)

    zip_file = os.path.join(benchmark_folder, "qf_lra.zip")
    zip_checksums = [
        'd0031a9e1799f78e72951aa4bacedaff7c0d027905e5de29b5980083b9c51138def165cc18fff205c1cdd0ef60d5d95cf179f0d82ec41ba489acf4383f3e783c']  # ,
    # '8aa31ada44bbb6705ce58f1f50870da4f3b2d2d27065f3c5c6a17bd484a4cb7eab0c1d55a8d78e48217e66c5b2d876c0708516fb8a383d1ea82a6d4f1278d476']
    qf_lra_folder = os.path.join(benchmark_folder, "QF_LRA")
    if not os.path.exists(qf_lra_folder) and not os.path.exists(zip_file):
        print("Downloading ZIP file to {}".format(zip_file))
        url = "http://smt-lib.loria.fr/zip/QF_LRA.zip"
        with urllib.request.urlopen(url) as response, open(zip_file, 'wb') as out_file:
            shutil.copyfileobj(response, out_file)
    if not os.path.exists(qf_lra_folder):
        print("Extracting ZIP file {}".format(zip_file))
        if checksum(zip_file) not in zip_checksums:
            fix_zip_file(zip_file)
            if checksum(zip_file) not in zip_checksums:
                raise RuntimeError("Corrupted file download")
        with zipfile.ZipFile(zip_file, 'r') as zip_ref:
            zip_ref.extractall(benchmark_folder)

    summary_file = os.path.join(benchmark_folder, "qf_lra_summary.pickle")

    if not os.path.exists(summary_file):
        with open(summary_file, "wb") as summary_file_ref:
            pickle.dump(dict(), summary_file_ref)
    with open(summary_file, "rb") as summary_file_ref:
        summary = pickle.load(summary_file_ref)

    cache_dir = os.path.join(benchmark_folder, "qf_lra_cache")
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    for filename in glob.glob("{}/**/*.smt*".format(qf_lra_folder), recursive=True):
        name = filename[filename.find("QF_LRA"):]
        if name not in summary:
            summary[name] = dict()

        cache_filename = "{}{}{}".format(cache_dir, os.path.sep, name)

        entry = summary[name]
        if "file_size" not in entry:
            entry["file_size"] = os.path.getsize(filename)

        if entry["file_size"] / 1024 > 100:
            continue

        density_filename = "{}.density".format(cache_filename)
        if not os.path.exists(os.path.dirname(density_filename)):
            os.makedirs(os.path.dirname(density_filename))

        domain, formula = None, None

        pysmt.environment.push_env()
        pysmt.environment.get_env().enable_infix_notation = True

        if not os.path.exists(density_filename):
            print("Importing {}".format(name))
            try:
                domain, formula = import_problem(filename)
                Density(domain, formula, smt.Real(1.0)).export_to(density_filename)
            except RuntimeError:
                print("Error")
                continue

        keys = ["real_variables_count", "bool_variables_count", "operators", "half_spaces"]
        if any(k not in entry for k in keys) and (domain is None or formula is None):
            print("Loading {}".format(name))
            density = Density.import_from(density_filename)
            domain = density.domain
            formula = density.support

        if "real_variables_count" not in entry:
            entry["real_variables_count"] = len(domain.real_vars)
        if "bool_variables_count" not in entry:
            entry["bool_variables_count"] = len(domain.bool_vars)
        if "operators" not in entry:
            entry["operators"] = OperatorWalker().find_operators(formula)
        if "half_spaces" not in entry:
            entry["half_spaces"] = HalfSpaceWalker().find_half_spaces(formula)

        pysmt.environment.pop_env()

    with open(summary_file, "wb") as summary_file_ref:
        pickle.dump(summary, summary_file_ref)


def select_benchmark_files(entry_filter, summary_file=None):
    summary_file = summary_file or get_summary_file()

    with open(summary_file, "rb") as summary_file_ref:
        summary = pickle.load(summary_file_ref)

    cache_dir = get_benchmark_cache_dir()
    for name, entry in summary.items():
        if entry_filter(entry):
            cache_filename = "{}{}{}".format(cache_dir, os.path.sep, name)
            density_filename = "{}.density".format(cache_filename)

            yield name, entry, density_filename


def benchmark_filter(entry):
    return "real_variables_count" in entry and entry["real_variables_count"] + entry["bool_variables_count"] <= 10 and \
           "=" not in entry["operators"]


def edit_summary(callback):
    with open(get_summary_file(), "rb") as summary_file_reference:
        summary = pickle.load(summary_file_reference)

    callback(summary)

    with open(get_summary_file(), "wb") as summary_file_reference:
        pickle.dump(summary, summary_file_reference)


def prepare_ratios():
    sample_count = 1000
    bounds_pool = [(-1, 1), (-10, 10), (-100, 100), (-1000, 1000)]
    ratios = dict()
    for name, entry, density_filename in select_benchmark_files(lambda e: "bounds" not in e and benchmark_filter(e)):
        print("Finding ratios for {}".format(name))
        pysmt.environment.push_env()
        pysmt.environment.get_env().enable_infix_notation = True

        density = Density.import_from(density_filename)
        domain = density.domain

        result_bounds = []
        result_ratios = []
        for bounds in itertools.product(*[bounds_pool for _ in range(len(domain.real_vars))]):
            var_bounds = dict(zip(domain.real_vars, bounds))
            restricted_domain = Domain(domain.variables, domain.var_types, var_bounds)
            samples = uniform(restricted_domain, sample_count)
            labels = evaluate(restricted_domain, density.support, samples)
            positive_count = sum(labels)
            if 0 < positive_count < sample_count:
                ratio = positive_count / sample_count
                result_bounds.append(var_bounds)
                result_ratios.append(ratio)

        ratios[name] = list(zip(result_bounds, result_ratios))
        print(name, result_ratios)

        pysmt.environment.pop_env()

    with open(get_summary_file(), "rb") as summary_file_reference:
        summary = pickle.load(summary_file_reference)

    for name, bounds in ratios.items():
        summary[name]["bounds"] = bounds

    with open(get_summary_file(), "wb") as summary_file_reference:
        pickle.dump(summary, summary_file_reference)


def prepare_samples(n, sample_size, reset):
    samples_dir = get_benchmark_samples_dir()

    seeds = [random.randint(0, 2 ** 32 - 1) for _ in range(n)]
    samples_dict = dict()

    def sample_filter(_entry):
        if "bounds" in _entry and benchmark_filter(_entry):
            if "samples" not in _entry["samples"]:
                return True
            else:
                return reset or any(
                    len([
                        s for s in _entry["samples"]
                        if s["sample_size"] == sample_size and s["bounds"] == _bounds[0]]
                    ) < n
                    for _bounds in _entry["bounds"] if 0.2 <= _bounds[1] <= 0.8
                )
        return False

    for name, entry, filename in select_benchmark_files(sample_filter):
        print("Creating samples for {}".format(name))
        pysmt.environment.push_env()
        pysmt.environment.get_env().enable_infix_notation = True

        density = Density.import_from(filename)
        samples_dict[name] = [] if reset else entry.get("samples", [])

        for i, (bounds, ratio) in enumerate(entry["bounds"]):
            if not (0.2 <= ratio <= 0.8):
                continue

            print(i, bounds, ratio)
            previous_samples = [] if reset else ([s for s in entry.get("samples", [])
                                                 if s["sample_size"] == sample_size and s["bounds"] == bounds])
            bounded_domain = Domain(density.domain.variables, density.domain.var_types, bounds)

            for j in range(n - len(previous_samples)):
                seed = seeds[j]
                samples_filename = "{}{}{}.{}.{}.{}.sample.npy".format(samples_dir, os.path.sep, name, sample_size,
                                                                       seed, i)
                labels_filename = "{}{}{}.{}.{}.{}.labels.npy".format(samples_dir, os.path.sep, name, sample_size, seed,
                                                                      i)

                if not os.path.exists(os.path.dirname(samples_filename)):
                    os.makedirs(os.path.dirname(samples_filename))

                random.seed(seed)
                np.random.seed(seed)
                samples = uniform(bounded_domain, sample_size)
                labels = evaluate(bounded_domain, density.support, samples)
                np.save(samples_filename, samples)
                np.save(labels_filename, labels)

                samples_dict[name].append({
                    "bounds": bounds,
                    "seed": seed,
                    "samples_filename": samples_filename,
                    "labels_filename": labels_filename,
                    "sample_size": sample_size
                })

        pysmt.environment.pop_env()

    def edit(summary):
        for _n, _s in samples_dict.items():
            summary[_n]["samples"] = _s

    edit_summary(edit)


def get_synthetic_db(directory, auto_dump=True):
    db_filename = os.path.join(directory, 'synthetic.db')
    return pickledb.load(db_filename, auto_dump)


def prepare_synthetic(input_directory, output_directory, runs, sample_size):
    seeds = [random.randint(0, 2 ** 32 - 1) for _ in range(runs)]

    db = get_synthetic_db(output_directory, True)
    os.makedirs(output_directory)
    for filename in glob.glob("{}/**/synthetics*.txt".format(input_directory), recursive=True):
        pysmt.environment.push_env()
        pysmt.environment.get_env().enable_infix_notation = True
        with open(filename) as file_reference:
            flat = json.load(file_reference)

        name = flat["synthetic_problem"]["problem"]["name"]
        print(name)

        if not db.exists(name):
            domain = import_domain(flat["synthetic_problem"]["problem"]["domain"])
            formula = nested_to_smt(flat["synthetic_problem"]["problem"]["theory"])
            Density(domain, formula, smt.Real(1.0)).export_to(os.path.join(output_directory, "{}.density".format(name)))
            entry = {
                "domain": export_domain(domain),
                "generation": {
                    "h": flat["synthetic_problem"]["half_space_count"],
                    "k": flat["synthetic_problem"]["formula_count"],
                    "l": flat["synthetic_problem"]["terms_per_formula"],
                    "structure": flat["synthetic_problem"]["cnf_or_dnf"],
                },
                "formula": smt_to_nested(formula),
                "samples": []
            }
        else:
            entry = dict(db.get(name))
            domain = import_domain(entry["domain"])
            formula = import_domain(entry["domain"])

        samples = entry.get("samples", [])
        matching_samples = []
        for sample in samples:
            if sample["sample_size"] == sample_size:
                matching_samples.append(sample)

        for i in range(runs - len(matching_samples)):
            seed = seeds[len(matching_samples) + i]
            samples_file = "{}.{}.{}.samples.npy".format(name, sample_size, seed)
            labels_file = "{}.{}.{}.labels.npy".format(name, sample_size, seed)
            np.random.seed(seed)
            data = uniform(domain, sample_size)
            np.save(os.path.join(output_directory, samples_file), data)
            labels = evaluate(domain, formula, data)
            np.save(os.path.join(output_directory, labels_file), labels)
            samples.append({
                "sample_size": sample_size,
                "seed": seed,
                "samples_file": samples_file,
                "labels_file": labels_file
            })

        entry["samples"] = samples
        db.set(name, entry)

        pysmt.environment.pop_env()



import argparse

import numpy as np
from pywmi.smt_print import pretty_print

from .learn import learn_benchmark, get_experiment, learn_synthetic
from .prepare import prepare_smt_lib_benchmark, prepare_ratios, prepare_samples, prepare_synthetic
from incal.learn import LearnOptions
from . import examples
from .analyze import analyze
from incal.util import analyze as show


def main():
    smt_lib_name = "smt-lib-benchmark"
    synthetic_name = "synthetic"
    parser = argparse.ArgumentParser(description="Interface with benchmark or synthetic data for experiments")

    parser.add_argument("source")
    parser.add_argument("--sample_size", type=int, default=None)
    parser.add_argument("--runs", type=int, default=None)
    parser.add_argument("--input_dir", type=str, default=None)
    parser.add_argument("--output_dir", type=str, default=None)
    parser.add_argument("--processes", type=int, default=None)
    parser.add_argument("--time_out", type=int, default=None)

    task_parsers = parser.add_subparsers(dest="task")
    prepare_parser = task_parsers.add_parser("prepare")
    prepare_parser.add_argument("--reset_samples", type=bool, default=False)
    learn_parser = task_parsers.add_parser("learn")
    analyze_parser = task_parsers.add_parser("analyze")
    analyze_parser.add_argument("--dirs", nargs="+", type=str)
    analyze_parser.add_argument("--res_path", type=str, default=None)

    show_parsers = analyze_parser.add_subparsers()
    show_parser = show_parsers.add_parser("show")
    show.add_arguments(show_parser)

    learn_options = LearnOptions()
    learn_options.add_arguments(learn_parser)

    args = parser.parse_args()
    if args.task == "prepare":
        if args.source == smt_lib_name:
            prepare_smt_lib_benchmark()
            prepare_ratios()
            prepare_samples(args.runs, args.sample_size, args.reset_samples)
        elif args.source == synthetic_name:
            prepare_synthetic(args.input_dir, args.output_dir, args.runs, args.sample_size)
    elif args.task == "learn":
        learn_options.parse_arguments(args)
        if args.source == smt_lib_name:
            learn_benchmark(args.runs, args.sample_size, args.processes, args.time_out, learn_options)
        elif args.source == synthetic_name:
            learn_synthetic(args.input_dir, args.output_dir, args.runs, args.sample_size, args.processes, args.time_out, learn_options)
        elif args.source.startswith("ex"):
            example_name = args.source.split(":", 1)[1]
            domain, formula = examples.get_by_name(example_name)
            np.random.seed(1)
            from pywmi.sample import uniform
            samples = uniform(domain, args.sample_size)
            from pywmi import evaluate
            labels = evaluate(domain, formula, samples)
            learn_options.set_value("domain", domain, False)
            learn_options.set_value("data", samples, False)
            learn_options.set_value("labels", labels, False)
            (formula, k, h), duration = learn_options.call(True)
            print("[{:.2f}s] Learned formula (k={}, h={}): {}".format(duration, k, h, pretty_print(formula)))
    elif args.task == "analyze":
        analyze(args.dirs, args.res_path, show.parse_args(args))



if __name__ == "__main__":
    main()



from pysmt.fnode import FNode
from pywmi import Domain
from typing import List


class Model(object):
    def __init__(self, domain: Domain, objective: FNode, constraints: List[FNode], minimize: bool=True, name=None):
        self.domain = domain
        self.objective = objective
        self.constraints = constraints
        self.minimize = minimize
        self.name = name



import glob
import json
import os

from typing import List

import numpy as np
import pickledb
from pywmi import RejectionEngine, nested_to_smt, import_domain
from pywmi.domain import Density, Domain

from .prepare import select_benchmark_files, benchmark_filter, get_synthetic_db
from incal.util.options import Experiment
from incal.util import analyze as show

from .learn import get_experiment

import pysmt.shortcuts as smt
import pysmt.environment


class Properties(object):
    bounds = dict()
    db = None

    @staticmethod
    def to_name(filename):
        return filename[filename.find("QF_LRA"):filename.find("smt2")+4]

    @staticmethod
    def to_sample_name(filename):
        return filename[filename.find("QF_LRA"):]

    @staticmethod
    def to_synthetic_name(filename):
        parts = os.path.basename(filename).split(".")
        return parts[0]

    @staticmethod
    def compute(experiments):
        Properties.db = pickledb.load('example.db', True)
        if Properties.db.exists("bounds"):
            Properties.bounds = Properties.db.get("bounds")
        else:
            used_names = {Properties.to_sample_name(e.parameters.original_values["data"]) for e in experiments}
            names_to_bounds = dict()
            summary_file = "remote_res/smt_lib_benchmark/qf_lra_summary.pickle"
            for name, entry, density_filename in select_benchmark_files(benchmark_filter, summary_file):
                if "samples" in entry:
                    for s in entry["samples"]:
                        name = Properties.to_sample_name(s["samples_filename"])
                        if name in used_names:
                            names_to_bounds[name] = s["bounds"]
            Properties.bounds = names_to_bounds
            Properties.db.set("bounds", Properties.bounds)

    @staticmethod
    def get_bound(experiment):
        return Properties.bounds[Properties.to_sample_name(experiment.parameters.original_values["data"])]

    @staticmethod
    def get_db_synthetic(experiment):
        return get_synthetic_db(os.path.dirname(experiment.parameters.original_values["domain"]))

    @staticmethod
    def original_k(experiment):
        db = Properties.get_db_synthetic(experiment)
        name = Properties.to_synthetic_name(experiment.imported_from_file)
        return db.get(name)["generation"]["k"]

    @staticmethod
    def original_h(experiment):
        db = Properties.get_db_synthetic(experiment)
        name = Properties.to_synthetic_name(experiment.imported_from_file)
        return db.get(name)["generation"]["h"]

    @staticmethod
    def original_l(experiment):
        db = Properties.get_db_synthetic(experiment)
        name = Properties.to_synthetic_name(experiment.imported_from_file)
        return db.get(name)["generation"]["l"]

    @staticmethod
    def executed(experiment):
        return 1 if experiment.results.duration is not None else 0

    @staticmethod
    def positive_ratio(experiment):
        labels = np.load(experiment.parameters.original_values["labels"])
        return sum(labels) / len(labels)

    @staticmethod
    def accuracy_approx(experiment):
        key = "accuracy_approx:{}".format(experiment.imported_from_file)
        if Properties.db.exists(key):
            return Properties.db.get(key)
        else:
            pysmt.environment.push_env()
            pysmt.environment.get_env().enable_infix_notation = True
            if os.path.basename(experiment.imported_from_file).startswith("synthetic"):
                db = Properties.get_db_synthetic(experiment)
                name = Properties.to_synthetic_name(experiment.imported_from_file)
                entry = db.get(name)
                domain = import_domain(json.loads(entry["domain"]))
                true_formula = nested_to_smt(entry["formula"])
            else:
                density = Density.import_from(experiment.parameters.original_values["domain"])
                domain = Domain(density.domain.variables, density.domain.var_types, Properties.get_bound(experiment))
                true_formula = density.support
            learned_formula = nested_to_smt(experiment.results.formula)
            engine = RejectionEngine(domain, smt.TRUE(), smt.Real(1.0), 100000)
            accuracy = engine.compute_probability(smt.Iff(true_formula, learned_formula))
            pysmt.environment.pop_env()
            print(accuracy)
            Properties.db.set(key, accuracy)
        return accuracy


def register_derived(experiment):
    experiment.register_derived("accuracy_approx", Properties.accuracy_approx)
    experiment.register_derived("original_h", Properties.original_h)
    experiment.register_derived("original_l", Properties.original_l)
    experiment.register_derived("original_k", Properties.original_k)
    experiment.register_derived("executed", Properties.executed)
    experiment.register_derived("pos_rate", Properties.positive_ratio)
    return experiment


def analyze(results_directories, res_path, show_args):
    experiments = []  # type: List[Experiment]
    for results_directory in results_directories:
        for filename in glob.glob("{}/**/*.result".format(results_directory), recursive=True):
            log_file = filename.replace(".result", ".log")
            if not os.path.exists(log_file):
                log_file = None
            experiment = get_experiment(res_path).load(filename)
            experiments.append(register_derived(experiment))

    Properties.compute(experiments)
    show.show(experiments, *show_args)



import random
import numpy as np

from autodora.experiment import Experiment, Parameter, Result, derived
from typing import Union

from incal.generator import SyntheticFormula
from incal import Formula
from incal.k_cnf_smt_learner import KCnfSmtLearner
from incal.learn import learn
from incal.violations.core import RandomViolationsStrategy


class LearnExperiment(Experiment):
    formula_filename = Parameter(
        str, None, "The filename of the (synthetic) formula file"
    )
    data_filename = Parameter(str, None, "The filename of the examples")
    labels_filename = Parameter(str, None, "The filename of the labels")

    learner = Parameter(str, None, "The type of learner (options: cnf.[mhv1])")
    initial = Parameter(str, "random.20", "The strategy to pick the initial examples")
    selection = Parameter(
        str, "random.10", "The strategy to pick the violating examples"
    )

    initial_k = Parameter(int, 1, "The initial number of constraints")
    initial_h = Parameter(int, 0, "The initial number of inequalities")

    cost_k = Parameter(float, 1, "The complexity cost of adding a constraints")
    cost_h = Parameter(float, 1, "The complexity cost of adding an inequality")

    seed = Parameter(int, None, "The seed to run the experiment with")
    log = Parameter(str, None, "The filename of the log file to log to")

    k = Result(int, None, "The number of constraints in the solution")
    h = Result(int, None, "The number of inequalities in the solution")
    learned_formula = Result(object, None, "The learned formula")

    @derived(cache=False)
    def formula(self) -> Union[Formula, SyntheticFormula]:
        return Formula.from_file(self["formula"])

    @derived(cache=True)
    def gen_k(self):
        raise NotImplementedError()

    @derived(cache=True)
    def gen_h(self):
        raise NotImplementedError()

    @derived(cache=True)
    def gen_l(self):
        raise NotImplementedError()

    def run_internal(self):
        random.seed(self["seed"])
        np.random.seed(self["seed"])

        formula = Formula.from_file(self["formula_filename"])
        data = np.load(self["data_filename"])
        labels = np.load(self["labels_filename"])

        learner_type, symmetries = self["learner"].split(".")
        assert learner_type == "cnf"

        def cnf_factory(k, h, selection_strategy):
            return KCnfSmtLearner(k, h, selection_strategy, symmetries=symmetries)

        initial_type, initial_count = self["initial"].split(".")
        assert initial_type == "random"

        def random_selection(indices):
            return random.sample(indices, int(initial_count))

        selection_type, selection_count = self["selection"].split(".")
        assert selection_type == "random"

        learn(
            formula.domain,
            data,
            labels,
            cnf_factory,
            random_selection,
            RandomViolationsStrategy(int(selection_count)),
            self["initial_k"],
            self["initial_h"],
            self["cost_k"],
            self["cost_h"],
            self["log"],
        )


LearnExperiment.enable_cli()



import heapq

import time

from .learner import NoFormulaFound


class ParameterFrontier(object):
    def __init__(self, w_k, w_h):
        self.c = lambda k, h: w_k * k + w_h * h
        self.pq = []
        self.tried = set()

    def push(self, k, h):
        if (k, h) not in self.tried:
            heapq.heappush(self.pq, (self.c(k, h), k, h))
            self.tried.add((k, h))

    def pop(self):
        c, k, h = heapq.heappop(self.pq)
        return k, h


def learn_bottom_up(data, labels, learn_f, w_k, w_h, init_k=1, init_h=0, max_k=None, max_h=None):
    """
    Learns a CNF(k, h) SMT formula phi using the learner encapsulated in init_learner such that
    C(k, h) = w_k * k + w_h * h is minimal.
    :param data: List of tuples of assignments and labels
    :param labels: Array of labels
    :param learn_f: Function called with data, k and h: learn_f(data, k, h)
    :param w_k: The weight assigned to k
    :param w_h: The weight assigned to h
    :param init_k:  The minimal value for k
    :param init_h:  The minimal value for h
    :param max_k:   The maximal value for k
    :param max_h:   The maximal value for h
    :return: A tuple containing: 1) the CNF(k, h) formula phi with minimal complexity C(k, h); 2) k; and 3) h
    """
    solution = None
    frontier = ParameterFrontier(w_k, w_h)
    frontier.push(init_k, init_h)
    i = 0
    while solution is None:
        i += 1
        k, h = frontier.pop()
        # print("Attempting to solve with k={} and h={}".format(k, h))
        start = time.time()
        try:
            solution = learn_f(data, labels, i, k, h)
            # print("Found solution after {:.2f}s".format(time.time() - start))
        except NoFormulaFound as e:
            data = e.data
            labels = e.labels
        if max_k is None or k + 1 <= max_k:
            frontier.push(k + 1, h)
        if max_h is None or h + 1 <= max_h:
            frontier.push(k, h + 1)
    return solution, k, h



import argparse
import filecmp
import fnmatch
import json

import os
import shutil

import migrate


def combine(output_dir, dirs, bias=None, prefix=None):
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    migrate.migrate_results(output_dir, bias)

    summary = os.path.join(output_dir, "problems.txt")
    if not os.path.isfile(summary):
        flat = {}
    else:
        with open(summary, "r") as f:
            flat = json.load(f)

    if prefix is not None:
        dirs = [str(prefix) + str(directory) for directory in dirs]

    for input_dir in dirs:
        migrate.migrate_results(input_dir, bias)
        input_summary = os.path.join(input_dir, "problems.txt")
        with open(input_summary, "r") as f:
            input_flat = json.load(f)
        for problem_id in input_flat:
            if problem_id not in flat:
                flat[problem_id] = {}
            for sample_size in input_flat[problem_id]:
                if sample_size not in flat[problem_id]:
                    flat[problem_id][sample_size] = input_flat[problem_id][sample_size]
                else:
                    raise RuntimeError("Attempting to overwrite sample size {} for problem {} from file {}"
                                       .format(sample_size, problem_id, input_summary))

    for input_dir in dirs:
        for input_file in os.listdir(input_dir):
            if fnmatch.fnmatch(input_file, '*.learning_log.txt'):
                old_file = os.path.join(input_dir, input_file)
                new_file = os.path.join(output_dir, input_file)
                if not os.path.isfile(new_file):
                    shutil.copy(old_file, new_file)
                else:
                    if not filecmp.cmp(old_file, new_file):
                        raise RuntimeError("Attempting to overwrite {} with {}".format(new_file, old_file))

    with open(summary, "w") as f:
        json.dump(flat, f)


def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument("output_dir")
    parser.add_argument("dirs", nargs="*")
    parser.add_argument("-b", "--bias", default=None, help="Specify the bias")
    parser.add_argument("-p", "--prefix", default=None, help="Specify the prefix for input dirs")
    parsed = parser.parse_args()
    combine(parsed.output_dir, parsed.dirs, parsed.bias, parsed.prefix)


if __name__ == "__main__":
    parse()



from __future__ import print_function

import pysmt.shortcuts as smt
from pysmt.typing import REAL, BOOL

from incremental_learner import IncrementalLearner


class KDnfSmtLearner(IncrementalLearner):
    def __init__(self, conjunction_count, half_space_count, selection_strategy, allow_negations=True):
        IncrementalLearner.__init__(self, "dnf_smt", selection_strategy)
        self.conjunction_count = conjunction_count
        self.half_space_count = half_space_count
        self.allow_negations = allow_negations

    def learn_partial(self, solver, domain, data, new_active_indices):
        # Constants
        n_b_original = len(domain.bool_vars)
        n_b = n_b_original * 2
        n_r = len(domain.real_vars)

        n_h_original = self.half_space_count if n_r > 0 else 0
        n_h = n_h_original * 2 if self.allow_negations else n_h_original

        n_c = self.conjunction_count
        n_d = len(data)

        real_features = [[row[v] for v in domain.real_vars] for row, _ in data]
        bool_features = [[row[v] for v in domain.bool_vars] for row, _ in data]
        labels = [row[1] for row in data]

        # Variables
        a_hr = [[smt.Symbol("a_hr[{}][{}]".format(h, r), REAL) for r in range(n_r)] for h in range(n_h_original)]
        b_h = [smt.Symbol("b_h[{}]".format(h), REAL) for h in range(n_h_original)]
        s_ch = [[smt.Symbol("s_ch[{}][{}]".format(c, h)) for h in range(n_h)] for c in range(n_c)]
        s_cb = [[smt.Symbol("s_cb[{}][{}]".format(c, b)) for b in range(n_b)] for c in range(n_c)]

        # Aux variables
        s_ih = [[smt.Symbol("s_ih[{}][{}]".format(i, h)) for h in range(n_h)] for i in range(n_d)]
        s_ic = [[smt.Symbol("s_ic[{}][{}]".format(i, c)) for c in range(n_c)] for i in range(n_d)]

        # Constraints
        for i in new_active_indices:
            x_r, x_b, label = real_features[i], bool_features[i], labels[i]

            for h in range(n_h_original):
                sum_coefficients = smt.Plus([a_hr[h][r] * smt.Real(x_r[r]) for r in range(n_r)])
                solver.add_assertion(smt.Iff(s_ih[i][h], sum_coefficients <= b_h[h]))

            for h in range(n_h_original, n_h):
                solver.add_assertion(smt.Iff(s_ih[i][h], ~s_ih[i][h - n_h_original]))

            for c in range(n_c):
                solver.add_assertion(smt.Iff(s_ic[i][c], smt.And(
                    [smt.TRUE()]
                    + [(~s_ch[c][h] | s_ih[i][h]) for h in range(n_h)]
                    + [~s_cb[c][b] for b in range(n_b_original) if not x_b[b]]
                    + [~s_cb[c][b] for b in range(n_b_original, n_b) if x_b[b - n_b_original]]
                )))

            if label:
                solver.add_assertion(smt.Or([s_ic[i][c] for c in range(n_c)]))
            else:
                solver.add_assertion(smt.And([~s_ic[i][c] for c in range(n_c)]))

        solver.solve()
        model = solver.get_model()

        x_vars = [domain.get_symbol(domain.real_vars[r]) for r in range(n_r)]
        half_spaces = [
            smt.Plus([model.get_value(a_hr[h][r]) * x_vars[r] for r in range(n_r)]) <= model.get_value(b_h[h])
            for h in range(n_h_original)
        ] + [
            smt.Plus([model.get_value(a_hr[h][r]) * x_vars[r] for r in range(n_r)]) > model.get_value(b_h[h])
            for h in range(n_h - n_h_original)
        ]

        b_vars = [domain.get_symbol(domain.bool_vars[b]) for b in range(n_b_original)]
        bool_literals = [b_vars[b] for b in range(n_b_original)]
        bool_literals += [~b_vars[b] for b in range(n_b - n_b_original)]

        conjunctions = [
            [half_spaces[h] for h in range(n_h) if model.get_py_value(s_ch[c][h])]
            + [bool_literals[b] for b in range(n_b) if model.get_py_value(s_cb[c][b])]
            for c in range(n_c)
        ]

        return smt.Or([smt.And(conjunction) for conjunction in conjunctions])


# class KDnfSmtLearner(Learner):
#     def __init__(self, max_hyperplanes, max_terms, allow_negations=True):
#         Learner.__init__(self)
#         self.max_hyperplanes = max_hyperplanes
#         self.max_terms = max_terms
#         self.allow_negations = allow_negations
#
#     def learn(self, domain, data, initial_indices=None):
#         # Constants
#         n_b_original = len(domain.bool_vars)
#         n_b = n_b_original * 2 if self.allow_negations else n_b_original
#
#         n_f = len(domain.real_vars)
#         n_h_original = self.max_hyperplanes if n_f > 0 else 0
#         n_h = n_h_original * 2 if self.allow_negations else n_h_original
#
#         n_c = self.max_terms
#         n_d = len(data)
#
#         real_features = [[Learner._convert(row[v]) for v in domain.real_vars] for row, _ in data]
#         bool_features = [[bool(row[v].constant_value()) for v in domain.bool_vars] for row, _ in data]
#         labels = [row[1] for row in data]
#
#         # Variables
#         a_hf = [[smt.Symbol("a_hf[{}][{}]".format(h, f), REAL) for f in range(n_f)] for h in range(n_h_original)]
#         b_h = [smt.Symbol("b_h[{}]".format(h), REAL) for h in range(n_h_original)]
#         s_ch = [[smt.Symbol("s_ch[{}][{}]".format(c, h)) for h in range(n_h)] for c in range(n_c)]
#         s_cb = [[smt.Symbol("s_cb[{}][{}]".format(c, b)) for b in range(n_b)] for c in range(n_c)]
#
#         # Aux variables
#         s_ih = [[smt.Symbol("s_ih[{}][{}]".format(i, h)) for h in range(n_h)] for i in range(n_d)]
#         s_ic = [[smt.Symbol("s_ic[{}][{}]".format(i, c)) for c in range(n_c)] for i in range(n_d)]
#
#         # Constraints
#         start = time.time()
#         active_indices = list(range(len(data))) if initial_indices is None else initial_indices
#         remaining = list(range(len(data)))  # list(sorted(set(range(len(data))) - set(active_indices)))
#
#         hyperplane_dnf = []
#
#         def check_model(_x):
#             _formula = smt.Or([smt.And(hyperplane_conjunct) for hyperplane_conjunct in hyperplane_dnf])
#             substitution = {_var: _x[str(_var)] for _var in x_vars + b_vars}
#             return _formula.substitute(substitution).simplify().is_true()
#
#         print("Starting solver with {} examples".format(len(active_indices)))
#
#         with smt.Solver() as solver:
#             while len(active_indices) > 0:
#                 remaining = list(sorted(set(remaining) - set(active_indices)))
#                 for i in active_indices:
#                     x, x_b, label = real_features[i], bool_features[i], labels[i]
#
#                     for h in range(n_h_original):
#                         sum_coefficients = smt.Plus([a_hf[h][f] * smt.Real(x[f]) for f in range(n_f)])
#                         solver.add_assertion(smt.Iff(s_ih[i][h], sum_coefficients <= b_h[h]))
#
#                     for h in range(n_h_original, n_h):
#                         solver.add_assertion(smt.Iff(s_ih[i][h], ~s_ih[i][h - n_h_original]))
#
#                     for c in range(n_c):
#                         solver.add_assertion(smt.Iff(s_ic[i][c], smt.And(
#                             [(~s_ch[c][h] | s_ih[i][h]) for h in range(n_h)]
#                             + [~s_cb[c][b] for b in range(n_b_original) if not x_b[b]]
#                             + [~s_cb[c][b] for b in range(n_b_original, n_b) if x_b[b - n_b_original]]
#                         )))
#
#                     if label:
#                         solver.add_assertion(smt.Or([s_ic[i][c] for c in range(n_c)]))
#                     else:
#                         solver.add_assertion(smt.And([~s_ic[i][c] for c in range(n_c)]))
#
#                 solver.solve()
#                 model = solver.get_model()
#
#                 x_vars = [domain.get_symbol(domain.variables[f]) for f in range(n_f)]
#                 hyperplanes = [
#                     smt.Plus([model.get_value(a_hf[h][f]) * x_vars[f] for f in range(n_f)]) <= model.get_value(b_h[h])
#                     for h in range(n_h_original)]
#                 hyperplanes += [
#                     smt.Plus([model.get_value(a_hf[h][f]) * x_vars[f] for f in range(n_f)]) > model.get_value(b_h[h])
#                     for h in range(n_h - n_h_original)]
#
#                 b_vars = [domain.get_symbol(domain.bool_vars[b]) for b in range(n_b_original)]
#                 bool_literals = [b_vars[b] for b in range(n_b_original)]
#                 bool_literals += [~b_vars[b - n_b_original] for b in range(n_b_original, n_b)]
#
#                 hyperplane_dnf = [
#                     [hyperplanes[h] for h in range(n_h) if model.get_py_value(s_ch[c][h])]
#                     + [bool_literals[b] for b in range(n_b) if model.get_py_value(s_cb[c][b])]
#                     for c in range(n_c)
#                 ]
#
#                 active_indices = [i for i in remaining if labels[i] != check_model(data[i][0])]
#                 print("Found model violating {} examples".format(len(active_indices)))
#
#         time_taken = time.time() - start
#         print("Took {:.2f}s".format(time_taken))
#         return hyperplane_dnf
#


from __future__ import print_function

from gurobipy.gurobipy import *


# noinspection PyPep8Naming
from pysmt.shortcuts import LE, Plus, Real, Times, And, Or

from learner import Learner


class OCTLearner(Learner):
    def __init__(self, N_min, D, alpha):
        self.N_min = N_min
        self.D = D
        self.alpha = alpha
        self.mu = 0.005

    def learn(self, domain, data):
        # Create a new model
        m = Model("OCT")

        print("Data")
        for row, l in data:
            print(*(["{}: {:.2f}".format(v, Learner._convert(row[v])) for v in domain.variables] + [l]))
        print()

        # Computed constants
        L_hat = OCTLearner._get_misclassification(data)

        print("Majority label misclassification cost", L_hat)

        p = len(domain.variables)
        K = 2
        n = len(data)
        leaf_count = 2 ** self.D
        inline_count = leaf_count - 1

        T = list(range(leaf_count + inline_count))
        # T_B = [t for t in T if t < inline_count]
        # T_L = [t for t in T if t >= inline_count]

        print("Nodes: {} internal, {} leaves".format(inline_count, leaf_count))

        Y_ik = [
            [1 if data[i][1] else -1, -1 if data[i][1] else 1]
            for i in range(n)
        ]

        A_L = [[] for _ in range(inline_count + leaf_count)]  # TODO
        A_R = [[] for _ in range(inline_count + leaf_count)]  # TODO

        def populate_ancestors(index):
            left = (index + 1) * 2 - 1
            right = (index + 1) * 2
            if right < inline_count + leaf_count:
                A_L[left] = A_L[index] + [index]
                A_R[left] = A_R[index]
                A_L[right] = A_L[index]
                A_R[right] = A_R[index] + [index]
                populate_ancestors(left)
                populate_ancestors(right)
        populate_ancestors(0)

        print("A_L: {}, A_R: {}".format(A_L, A_R))

        x_ij = [[Learner._convert(row[v]) for v in domain.variables] for row, _ in data]
        p_t = [None if i == 0 else int((i + 1) / 2) - 1 for i in range(inline_count)]

        # Variables
        # x = m.addVar(vtype=GRB.BINARY, name="x")
        L_t = [m.addVar(vtype=GRB.INTEGER, name="L_{}".format(i)) for i in range(leaf_count)]
        s_jt = [
            [m.addVar(vtype=GRB.BINARY, name="s{}_{}".format(j, i)) for i in range(inline_count)]
            for j in range(p)
        ]
        N_t = [m.addVar(vtype=GRB.INTEGER, name="N_{}".format(i)) for i in range(leaf_count)]
        N_kt = [
            [m.addVar(vtype=GRB.INTEGER, name="N_{}_{}".format(k, i)) for i in range(leaf_count)]
            for k in range(K)
        ]
        c_kt = [
            [m.addVar(vtype=GRB.BINARY, name="c_{}_{}".format(k, i)) for i in range(leaf_count)]
            for k in range(K)
        ]
        z_it = [
            [m.addVar(vtype=GRB.BINARY, name="z_{}_{}".format(i, t)) for t in range(leaf_count)]
            for i in range(n)
        ]
        l_t = [m.addVar(vtype=GRB.BINARY, name="l_{}".format(t)) for t in range(leaf_count)]
        a_jt = [
            [m.addVar(vtype=GRB.CONTINUOUS, lb=-1, ub=1, name="a_{}_{}".format(j, t)) for t in range(inline_count)]
            for j in range(p)
        ]
        a_hat_jt = [
            [m.addVar(vtype=GRB.CONTINUOUS, name="a_hat_{}_{}".format(j, t)) for t in range(inline_count)]
            for j in range(p)
        ]
        b_t = [m.addVar(vtype=GRB.CONTINUOUS, name="b_{}".format(t)) for t in range(inline_count)]
        d_t = [m.addVar(vtype=GRB.BINARY, name="d_{}".format(t)) for t in range(inline_count)]
        print()

        # Set objective
        m.setObjective(1.0 / L_hat * sum(L_t) + self.alpha * sum(sum(entry) for entry in s_jt), GRB.MINIMIZE)

        # Add constraint: L_t >= N_t - N_kt - n(1 - c_kt) forall k = 1..K, t in T_L
        for k in range(K):
            for t in range(leaf_count):
                name = "L_{t} >= N_{t} - N_{k}_{t} - n(1 - c_{k}_{t})".format(k=k, t=t)
                m.addConstr(L_t[t] >= N_t[t] - N_kt[k][t] - n * (1 - c_kt[k][t]), name)
                print(name)
        print()

        # Add constraint: L_t <= N_t - N_kt + n * c_kt forall k = 1..K, t in T_L
        for k in range(K):
            for t in range(leaf_count):
                name = "L_{t} <= N_{t} - N_{k}_{t} + n * c_{k}_{t}".format(k=k, t=t)
                m.addConstr(L_t[t] <= N_t[t] - N_kt[k][t] + n * c_kt[k][t], name)
                print(name)
        print()

        # Add constraint: L_t >= 0 forall t in T_l
        for t in range(leaf_count):
            name = "L_{t} >= 0".format(t=t)
            m.addConstr(L_t[t] >= 0, name)
            print(name)
        print()

        # Add constraint: N_kt = 1/2 * Sum_i = 1..n (1 + Y_ik) * z_it forall k = 1..K, t in T_L
        for k in range(K):
            for t in range(leaf_count):
                name = "N_{k}_{t} = 1/2 * SUM_(i = 0..{n}) (1 + Y_i_{k}) * z_i_{t}".format(k=k, t=t, n=n-1)
                m.addConstr(N_kt[k][t] == sum((1 + Y_ik[i][k]) * z_it[i][t] for i in range(n)) / 2, name)
                print(name)
        print()

        # Add constraint: N_t = Sum_i = 1..n z_it forall t in T_L
        for t in range(leaf_count):
            name = "N_{t} = SUM_(i = 0..{n}) z_i_{t}".format(t=t, n=n-1)
            m.addConstr(N_t[t] == sum(z_it[i][t] for i in range(n)), name)
            print(name)
        print()

        # Add constraint: Sum_k = 1..K c_kt = l_t forall t in T_L
        for t in range(leaf_count):
            # print(sum([c_kt[k][t] for k in range(K)]))
            name = "SUM_(k = 0..{K}) c_k_{t} = l_{t}".format(t=t, K=K-1)
            m.addConstr(sum(c_kt[k][t] for k in range(K)) == l_t[t], name)
            print(name)
        print()

        # Add constraint: Sum_j = 1..p (a_jm^T * x_ji) + mu <= b_m + (2 + mu) * (1 - z_it)
        # forall i in 1..n, t in T_B, m in A_L(t)
        for i in range(n):
            for t in range(leaf_count):
                for _m in A_L[inline_count + t]:
                    vector_product = sum(a_jt[j][_m] * x_ij[i][j] for j in range(p))
                    name = "SUM_(j = 0..{p}) (a_j_{m}^T * x_j_{i}) + mu <= b_{m} + (2 + mu) * (1 - z_{i}_{t})"\
                        .format(i=i, t=t, m=_m, p=p-1)
                    m.addConstr(vector_product + self.mu <= b_t[_m] + (2 + self.mu) * (1 - z_it[i][t]), name)
                    print(name)
        print()

        # TODO, both times 1 - z_it??
        # Add constraint: Sum_j = 1..p a_jm^T * x_ji >= b_m - 2 * (1 - z_it) forall i in 1..n, t in T_B, m in A_R(t)
        for i in range(n):
            for t in range(leaf_count):
                for _m in A_R[inline_count + t]:
                    name = "SUM_(j = 0..{p}) a_j_{m}^T * x_j_{i} >= b_{m} - 2 * (1 - z_{i}_{t})"\
                        .format(i=i, t=t, m=_m, p=p-1)
                    vector_product = sum(a_jt[j][_m] * x_ij[i][j] for j in range(p))
                    m.addConstr(vector_product >= b_t[_m] - 2 * (1 - z_it[i][t]), name)
                    print(name)
        print()

        # Add constraint: Sum_t in T_L z_it = 1 forall i = 1..n
        for i in range(n):
            name = "SUM_(t in {T_L}) z_{i}_t = 1".format(i=i, T_L=list(range(leaf_count)))
            m.addConstr(sum(z_it[i][t] for t in range(leaf_count)) == 1, name)
            print(name)
        print()

        # Add constraint: z_it <= l_t forall i = 1..n, t in T_L
        for i in range(n):
            for t in range(leaf_count):
                name = "z_{i}_{t} <= l_{t}".format(i=i, t=t)
                m.addConstr(z_it[i][t] <= l_t[t], name)
                print(name)
        print()

        # Add constraint: Sum_i = 1..n z_it >= N_min * l_t forall t in T_L
        for t in range(leaf_count):
            name = "SUM_(i = 0..{n}) z_i_{t} >= N_min * l_{t}".format(t=t, n=n-1)
            m.addConstr(sum(z_it[i][t] for i in range(n)) >= self.N_min * l_t[t], name)
            print(name)
        print()

        # Add constraint: Sum_j = 1..p a_hat_jt <= d_t forall t in T_B
        for t in range(inline_count):
            name = "SUM_(j = 0..{p}) a_hat_j_{t} <= d_{t}".format(t=t, p=p-1)
            m.addConstr(sum(a_hat_jt[j][t] for j in range(p)) <= d_t[t], name)
            print(name)
        print()

        # Add constraint: a_hat_jt >= a_jt forall j = 1..p, t in T_B
        for j in range(p):
            for t in range(inline_count):
                name = "a_hat_{j}_{t} >= a_{j}_{t}".format(j=j, t=t)
                m.addConstr(a_hat_jt[j][t] >= a_jt[j][t], name)
                print(name)
            print()

        # Add constraint: a_hat_jt >= -a_jt forall j = 1..p, t in T_B
        for j in range(p):
            for t in range(inline_count):
                name = "a_hat_{j}_{t} >= -a_{j}_{t}".format(j=j, t=t)
                m.addConstr(a_hat_jt[j][t] >= -a_jt[j][t], name)
                print(name)
            print()

        # Add constraint: -s_jt <= a_jt <= s_jt forall j = 1..p, t in T_B
        for j in range(p):
            for t in range(inline_count):
                name1 = "-s_{j}_{t} <= a_{j}_{t}".format(j=j, t=t)
                m.addConstr(-s_jt[j][t] <= a_jt[j][t], name1)
                name2 = "s_{j}_{t} >= a_{j}_{t}".format(j=j, t=t)
                m.addConstr(s_jt[j][t] >= a_jt[j][t], name2)
                print(name1)
                print(name2)
        print()

        # Add constraint: s_jt <= d_t forall j = 1..p, t in T_B
        for j in range(p):
            for t in range(inline_count):
                name = "s_{j}_{t} <= d_{t}".format(j=j, t=t)
                m.addConstr(s_jt[j][t] <= d_t[t], name)
                print(name)
        print()

        # Add constraint: Sum_j = 1..p s_jt >= d_t forall t in T_B
        for t in range(inline_count):
            name = "SUM_(j = 0..{p}) s_j_{t} >= d_{t}".format(t=t, p=p-1)
            m.addConstr(sum(s_jt[j][t] for j in range(p)) >= d_t[t], name)
            print(name)
        print()

        # Add constraint: -d_t <= b_t <= d_t forall t in T_B
        for t in range(inline_count):
            name1 = "-d_{t} <= b_{t}".format(t=t)
            m.addConstr(-d_t[t] <= b_t[t], name1)
            name2 = "d_{t} >= b_{t}".format(t=t)
            m.addConstr(d_t[t] >= b_t[t], name2)
            print(name1)
            print(name2)
        print()

        # Add constraint: d_t <= d_p(t) forall t in T_B \ {1}
        for t in range(1, inline_count):
            name = "d_{t} <= d_{p_t}".format(t=t, p_t=p_t[t])
            m.addConstr(d_t[t] <= d_t[p_t[t]], name)
            print(name)
        print()

        # Hack example
        # m.addConstr(N_t[0] == 2)
        # m.addConstr(N_t[1] == 1)

        # m.addConstr(z_it[0][0] == 1)
        # m.addConstr(z_it[1][0] == 1)
        # m.addConstr(z_it[2][0] == 0)
        # m.addConstr(z_it[0][1] == 0)
        # m.addConstr(z_it[1][1] == 0)
        # m.addConstr(z_it[2][1] == 1)

        # m.addConstr(c_kt[0][0] == 1)
        # m.addConstr(c_kt[1][0] == 0)
        # m.addConstr(c_kt[0][1] == 0)
        # m.addConstr(c_kt[1][1] == 1)

        # m.addConstr(a_jt[0][0] == 1)
        # m.addConstr(b_t[0] == 0.6)

        m.optimize()

        for v in m.getVars():
            print(v.varName, v.x)

        print('Obj:', m.objVal)

        paths = []

        def populate_paths(index, prefix=None):
            if prefix is None:
                prefix = []
            left = (index + 1) * 2 - 1
            right = (index + 1) * 2
            if right < inline_count and d_t[index].x > 0.5:
                coefficients = [Real(float(a_jt[j][index].x)) for j in range(p)]
                linear_sum = Plus([Times(c, domain.get_symbol(v)) for c, v in zip(coefficients, domain.variables)])
                constant = Real(float(b_t[index].x))
                node = prefix + [LE(linear_sum, constant)]
                populate_paths(left, node)
                populate_paths(right, node)
            else:
                paths.append(prefix)
        populate_paths(0)
        return paths




from __future__ import print_function

import itertools

import numpy as np
from pysmt.shortcuts import Plus, Real, Times, LE, GE
from pysmt.typing import REAL, BOOL

from learner import Learner


class KDNFLogicLearner(Learner):
    def __init__(self, k):
        Learner.__init__(self)
        self.k = k

    def learn(self, domain, data, border_indices):
        positive_indices = [i for i in range(len(data)) if data[i][1]]
        real_vars = [v for v in domain.variables if domain.var_types[v] == REAL]
        bool_vars = [v for v in domain.variables if domain.var_types[v] == BOOL]
        d = len(real_vars)
        hyperplanes = []
        for indices in itertools.combinations(positive_indices, d):
            print(indices)
            hyperplanes.append(Learner.fit_hyperplane(domain, [data[i][0] for i in indices]))
        boolean_data = []
        for i in range(len(data)):
            row = []
            for v in bool_vars:
                row.append(data[i][0][v].constant_value())
            boolean_data.append(row)
        hyperplanes_smt = []
        for a, c in hyperplanes:
            lhs_smt = Plus(Times(Real(float(a[j])), domain.get_symbol(real_vars[j])) for j in range(d))
            hyperplanes_smt.append(LE(lhs_smt, Real(c)))
            lhs_smt = Plus(Times(Real(-float(a[j])), domain.get_symbol(real_vars[j])) for j in range(d))
            hyperplanes_smt.append(LE(lhs_smt, Real(-c)))
            for i in range(len(data)):
                lhs = 0
                for j in range(d):
                    lhs += float(a[j]) * float(data[i][0][real_vars[j]].constant_value())
                boolean_data[i].append(lhs <= c)
                boolean_data[i].append(lhs >= c)
        print(boolean_data)
        # logical_dnf_indices = [[i] for i in range(len(boolean_data[0]))]
        logical_dnf_indices = self.learn_logical(boolean_data, [row[1] for row in data])
        logical_dnf = [
            [domain.get_symbol(bool_vars[i]) if i < len(bool_vars) else
             hyperplanes_smt[i - len(bool_vars)] for i in conj_indices]
            for conj_indices in logical_dnf_indices
        ]
        print(logical_dnf)
        return logical_dnf

    def learn_logical(self, boolean_data, labels):
        conjunctions = []
        for k in range(1, self.k + 1):
            for features in itertools.combinations(list(range(len(boolean_data))), k):
                accept = True
                for entry, label in zip(boolean_data, labels):
                    if not label and all(entry[j] for j in features):
                        accept = False
                        break
                if accept:
                    conjunctions.append(features)
        return conjunctions


class GreedyMaxRuleLearner(KDNFLogicLearner):
    def __init__(self, max_literals):
        KDNFLogicLearner.__init__(self, max_literals)

    def learn_logical(self, boolean_data, labels):
        attributes = np.matrix(boolean_data)
        examples = attributes.shape[0]
        features = attributes.shape[1]
        conjunctions = []
        counts = np.sum(attributes, axis=0).A1
        print(examples, features, counts.shape)

        return []


class GreedyLogicDNFLearner(KDNFLogicLearner):
    def __init__(self, max_terms, max_literals):
        KDNFLogicLearner.__init__(self, max_literals)
        self.max_terms = max_terms

    @property
    def max_literals(self):
        return self.k

    def learn_logical(self, boolean_data, labels):
        attributes = np.matrix(boolean_data)
        examples = attributes.shape[0]
        features = attributes.shape[1]
        conjunctions = []
        counts = np.sum(attributes, axis=0).A1
        print(counts[0])

        for i in range(self.max_terms):
            lb = 0
            ub = examples
            candidates = [([], examples)]
            new_candidates = []
            while len(candidates) > 0:
                for pattern, count in candidates:
                    start_index = 0 if len(pattern) == 0 else max(pattern) + 1
                    covered = [i for i in range(examples) if all(attributes[i, j] for j in pattern)]
                    pos_covered = [i for i in covered if labels[i]]
                    neg_covered = [i for i in covered if not labels[i]]

                    for j in range(start_index, self.max_literals):
                        if counts[j] > lb:
                            pass
                for j in range(self.max_literals):
                    for features in itertools.combinations(list(range(len(boolean_data))), self.k):
                        accept = True
                        for entry, label in zip(boolean_data, labels):
                            if not label and all(entry[j] for j in features):
                                accept = False
                                break
                        if accept:
                            conjunctions.append(features)
        return conjunctions



from __future__ import print_function, division

import fnmatch
import hashlib
import json

import os
import random

import itertools
import pysmt.shortcuts as smt
import re

import time

import generator
import inc_logging
import parse
import problem
from incremental_learner import RandomViolationsStrategy
from k_cnf_smt_learner import KCnfSmtLearner
from k_dnf_smt_learner import KDnfSmtLearner
from parameter_free_learner import learn_bottom_up
from smt_print import pretty_print
from smt_walk import SmtWalker


class OperatorWalker(SmtWalker):
    def __init__(self):
        self.operators = set()

    def walk_and(self, args):
        self.operators.add("&")
        self.walk_smt_multiple(args)

    def walk_or(self, args):
        self.operators.add("|")
        self.walk_smt_multiple(args)

    def walk_plus(self, args):
        self.operators.add("+")
        self.walk_smt_multiple(args)

    def walk_minus(self, left, right):
        self.operators.add("-")
        self.walk_smt_multiple([left, right])

    def walk_times(self, args):
        self.operators.add("*")
        self.walk_smt_multiple(args)

    def walk_not(self, argument):
        self.operators.add("~")
        self.walk_smt_multiple([argument])

    def walk_ite(self, if_arg, then_arg, else_arg):
        self.operators.add("ite")
        self.walk_smt_multiple([if_arg, then_arg, else_arg])

    def walk_pow(self, base, exponent):
        self.operators.add("^")
        self.walk_smt_multiple([base, exponent])

    def walk_lte(self, left, right):
        self.operators.add("<=")
        self.walk_smt_multiple([left, right])

    def walk_lt(self, left, right):
        self.operators.add("<")
        self.walk_smt_multiple([left, right])

    def walk_equals(self, left, right):
        self.operators.add("=")
        self.walk_smt_multiple([left, right])

    def walk_symbol(self, name, v_type):
        pass

    def walk_constant(self, value, v_type):
        pass

    def find_operators(self, formula):
        self.walk_smt(formula)
        return list(self.operators)


class HalfSpaceWalker(SmtWalker):
    def __init__(self):
        self.half_spaces = set()

    def walk_and(self, args):
        self.walk_smt_multiple(args)

    def walk_or(self, args):
        self.walk_smt_multiple(args)

    def walk_plus(self, args):
        self.walk_smt_multiple(args)

    def walk_minus(self, left, right):
        self.walk_smt_multiple([left, right])

    def walk_times(self, args):
        self.walk_smt_multiple(args)

    def walk_not(self, argument):
        self.walk_smt_multiple([argument])

    def walk_ite(self, if_arg, then_arg, else_arg):
        self.walk_smt_multiple([if_arg, then_arg, else_arg])

    def walk_pow(self, base, exponent):
        self.walk_smt_multiple([base, exponent])

    def walk_lte(self, left, right):
        self.half_spaces.add(parse.smt_to_nested(left <= right))

    def walk_lt(self, left, right):
        self.half_spaces.add(parse.smt_to_nested(left < right))

    def walk_equals(self, left, right):
        self.walk_smt_multiple([left, right])

    def walk_symbol(self, name, v_type):
        pass

    def walk_constant(self, value, v_type):
        pass

    def find_half_spaces(self, formula):
        self.walk_smt(formula)
        return list(self.half_spaces)


def import_problem(name, filename):
    target_formula = smt.read_smtlib(filename)

    variables = target_formula.get_free_variables()
    var_names = [str(v) for v in variables]
    var_types = {str(v): v.symbol_type() for v in variables}
    var_domains = {str(v): (0, 200) for v in variables}  # TODO This is a hack

    domain = problem.Domain(var_names, var_types, var_domains)
    target_problem = problem.Problem(domain, target_formula, name)

    return target_problem


def get_cache_name():
    cache = os.path.join(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "demo"), "cache")
    return os.path.join(cache, "summary.json")


def get_problem_file(problem_id):
    return os.path.join(os.path.dirname(get_cache_name()), "problems", "{}.txt".format(problem_id))


def load():
    name = get_cache_name()
    if not os.path.exists(name):
        with open(name, "w") as f:
            json.dump({"files": {}}, f)

    with open(name, "r") as f:
        flat = json.load(f)
    return flat


def dump(flat):
    name = get_cache_name()
    temp_name = "{}.tmp".format(name)
    with open(temp_name, "w") as f:
        json.dump(flat, f)
    os.rename(temp_name, name)


def scan(full_dir, root_dir):
    matches = []

    for root, dir_names, file_names in os.walk(full_dir):
        for filename in fnmatch.filter(file_names, '*.smt2'):
            full_path = os.path.abspath(os.path.join(root, filename))
            matches.append(re.match("{r}{s}(.*)".format(s=os.path.sep, r=root_dir), full_path).group(1))

    for match in matches:
        print(match)

    flat = load()
    problems = flat["files"]
    counter = 0
    for match in matches:
        if match not in problems:
            print("Importing {}".format(match))
            smt_file = os.path.join(root_dir, match)
            if os.path.getsize(smt_file) / 1024 <= 100:
                imported_problem = import_problem(match, smt_file)
                problems[match] = {
                    "loaded": True,
                    "var_count": len(imported_problem.domain.variables),
                    "file_size": os.path.getsize(match)
                }
                smt.reset_env()
            else:
                problems[match] = {"loaded": False, "reason": "size", "file_size": os.path.getsize(match)}
            counter += 1
            if counter >= 1:
                dump(flat)
                counter = 0

    if counter != 0:
        dump(flat)


def has_equals(_props):
    return "=" in [str(o) for o in _props["operators"]]


def has_conjunctions(_props):
    return "&" in [str(o) for o in _props["operators"]]


def has_disjunctions(_props):
    return "|" in [str(o) for o in _props["operators"]]


def has_connectives(_props):
    return has_conjunctions(_props) or has_disjunctions(_props)


def var_count(_props):
    return _props["var_count"]


def half_space_count(_props):
    return len(_props["half_spaces"])


def file_size(_props):
    return _props["file_size"]


def analyze(root_dir):
    flat = load()
    files = flat["files"]
    if "lookup" not in flat:
        flat["lookup"] = dict()
    lookup = flat["lookup"]

    print(*["mode", "var_count", "half_space_count", "file_size", "name"], sep="\t")
    for name, properties in files.items():
        smt_file = os.path.join(root_dir, name)
        if properties["loaded"] and properties["var_count"] < 10:
            change = False
            if "id" not in properties:
                properties["id"] = hashlib.sha256(smt_file).hexdigest()
                change = True

            if properties["id"] not in lookup:
                lookup[properties["id"]] = name
                change = True

            problem_file = get_problem_file(properties["id"])
            if not os.path.isfile(problem_file):
                with open(problem_file, "w") as f:
                    print(problem.export_problem(import_problem(name, smt_file)), file=f)

            with open(problem_file, "r") as f:
                target_problem = problem.import_problem(json.load(f))

            if "operators" not in properties:
                properties["operators"] = OperatorWalker().find_operators(target_problem.theory)
                change = True

            if "half_spaces" not in properties:
                properties["half_spaces"] = HalfSpaceWalker().find_half_spaces(target_problem.theory)
                change = True

            if change:
                dump(flat)

            if not has_equals(properties) and has_connectives(properties):
                mode = "D" if has_disjunctions(properties) else "C"
                information = [mode, var_count(properties), half_space_count(properties), file_size(properties), name]
                print(*[str(i) for i in information], sep="\t")


def learn_formula(problem_id, domain, h, data, seed, log_dir, learn_all=False, learn_dnf=False):
    initial_size = 20
    violations_size = 10
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    def learn_inc(_data, _k, _h):
        if learn_all is not None and learn_all:
            initial_indices = list(range(len(data)))
        else:
            initial_indices = random.sample(list(range(len(data))), initial_size)
        violations_strategy = RandomViolationsStrategy(violations_size)
        if learn_dnf is not None and learn_dnf:
            learner = KDnfSmtLearner(_k, _h, violations_strategy)
        else:
            learner = KCnfSmtLearner(_k, _h, violations_strategy)
        log_file = os.path.join(log_dir, "{}_{}_{}_{}_{}.learning_log.txt".format(problem_id, len(data), seed, _k, _h))
        learner.add_observer(inc_logging.LoggingObserver(log_file, seed, True, violations_strategy))
        learned_theory = learner.learn(domain, data, initial_indices)
        # learned_theory = Or(*[And(*planes) for planes in hyperplane_dnf])
        print("Learned theory:\n{}".format(pretty_print(learned_theory)))
        return learned_theory

    phi, k, h = learn_bottom_up(data, learn_inc, 1, 1)

    overview = os.path.join(log_dir, "problems.txt")
    if not os.path.isfile(overview):
        flat = {}
    else:
        with open(overview, "r") as f:
            flat = json.load(f)
    if problem_id not in flat:
        flat[problem_id] = {}
    flat[problem_id][len(data)] = {"k": k, "h": h, "seed": seed, "bias": "dnf" if learn_dnf else "cnf"}
    with open(overview, "w") as f:
        json.dump(flat, f)


def learn(sample_count, log_dir, learn_all=False, learn_dnf=False):
    flat = load()
    files = flat["files"]
    ratio_dict = flat["ratios"]
    seed = hash(time.time())
    random.seed(seed)
    for name, props in files.items():
        if props["loaded"] and props["var_count"] < 10 and not has_equals(props) and has_disjunctions(props) and \
                ratio_dict[name]["finite"] and 0.2 <= ratio_dict[name]["ratio"] <= 0.8:
            with open(get_problem_file(props["id"]), "r") as f:
                target_problem = problem.import_problem(json.load(f))
            adapted_problem = adapt_domain_multiple(target_problem, ratio_dict[name]["bounds"])
            samples = generator.get_problem_samples(adapted_problem, sample_count, 1)
            domain = adapted_problem.domain
            learn_formula(props["id"], domain, len(props["half_spaces"]), samples, seed, log_dir, learn_all, learn_dnf)
            print(props["id"], name)


def adapt_domain(target_problem, lb, ub):
    domain = target_problem.domain
    var_domains = {}
    for v in domain.variables:
        var_domains[v] = (lb, ub)
    adapted_domain = problem.Domain(domain.variables, domain.var_types, var_domains)
    return problem.Problem(adapted_domain, target_problem.theory, target_problem.name)


def adapt_domain_multiple(target_problem, new_bounds):
    domain = target_problem.domain
    adapted_domain = problem.Domain(domain.variables, domain.var_types, new_bounds)
    return problem.Problem(adapted_domain, target_problem.theory, target_problem.name)


def ratios():
    flat = load()
    files = flat["files"]
    if "ratios" not in flat:
        flat["ratios"] = dict()
    ratio_dict = flat["ratios"]

    sample_count = 100
    for name, props in files.items():
        if props["loaded"] and props["var_count"] < 10 and not has_equals(props) and has_disjunctions(props):
            with open(get_problem_file(props["id"]), "r") as f:
                target_problem = problem.import_problem(json.load(f))

            bounds_pool = [(-1, 1), (-10, 10), (-100, 100), (-1000, 1000)]
            domain = target_problem.domain
            result = {"finite": False, "samples": sample_count}
            for bounds in itertools.product(*[bounds_pool for _ in range(len(domain.real_vars))]):
                var_bounds = dict(zip(domain.real_vars, bounds))
                current_problem = adapt_domain_multiple(target_problem, var_bounds)
                samples = generator.get_problem_samples(current_problem, sample_count, 1)
                positive_count = len([x for x, y in samples if y])
                is_finite = positive_count not in [0, sample_count]
                if is_finite:
                    ratio = positive_count / sample_count
                    if not result["finite"] or abs(ratio - 0.5) < abs(result["ratio"] - 0.5):
                        result["finite"] = True
                        result["ratio"] = ratio
                        result["bounds"] = var_bounds
                    if abs(ratio - 0.5) <= 0.1:
                        break
            print(result)
            ratio_dict[name] = result
    dump(flat)


def summarize(results_dir, output_type):
    results_file = os.path.join(results_dir, "problems.txt")
    with open(results_file, "r") as f:
        results_flat = json.load(f)

    overview = load()
    lookup = overview["lookup"]

    simple_output = output_type is None

    if simple_output:
        print("name", "sample_size", "total_duration", sep="\t")

    unique_names = set()
    unique_sample_sizes = set()
    duration_table = dict()
    k_table = dict()
    h_table = dict()

    for problem_id in results_flat:
        name = lookup.get(problem_id, problem_id)
        unique_names.add(name)
        for sample_size in results_flat[problem_id]:
            unique_sample_sizes.add(sample_size)
            timed_out = results_flat[problem_id][sample_size].get("time_out", False)
            seed, k, h = (results_flat[problem_id][sample_size][v] for v in ["seed", "k", "h"])
            log_file = "{}_{}_{}_{}_{}.learning_log.txt".format(problem_id, sample_size, seed, k, h)
            log_file_full = os.path.join(results_dir, log_file)
            if not timed_out:
                durations = []
                with open(log_file_full, "r") as f:
                    for line in f:
                        flat = json.loads(line)
                        if flat["type"] == "update":
                            durations.append(flat["selection_time"] + flat["solving_time"])
                duration_table[(name, sample_size)] = sum(durations)
                if simple_output:
                    print(name, sample_size, sum(durations), sep="\t")
            else:
                duration_table[(name, sample_size)] = "({})".format(results_flat[problem_id][sample_size]["time_limit"])
                if simple_output:
                    print(name, sample_size, None, sep="\t")
            k_table[(name, sample_size)] = k
            h_table[(name, sample_size)] = h

    names = list(sorted(unique_names))
    sample_sizes = list(sorted(unique_sample_sizes, key=int))

    if output_type in ["time", "k", "h"]:
        print("", *sample_sizes, sep="\t")

    if output_type == "time":
        for name in names:
            print(name, *[duration_table.get((name, sample_size), "") for sample_size in sample_sizes], sep="\t")

    if output_type == "k":
        for name in names:
            print(name, *[k_table.get((name, sample_size), "") for sample_size in sample_sizes], sep="\t")

    if output_type == "h":
        for name in names:
            print(name, *[h_table.get((name, sample_size), "") for sample_size in sample_sizes], sep="\t")


def load_results(results_dir):
    results_file = os.path.join(results_dir, "problems.txt")
    with open(results_file, "r") as f:
        results_flat = json.load(f)
    return results_flat


def dump_results(results_flat, results_dir):
    results_file = os.path.join(results_dir, "problems.txt")
    with open(results_file, "w") as f:
        json.dump(results_flat, f)


def get_log_messages(results_dir, config, p_id=None, samples=None):
    id_key = "problem_id" if "problem_id" in config else "id"
    sample_key = "sample_size" if "sample_size" in config else "samples"
    problem_id = config[id_key] if p_id is None else p_id
    sample_size = config[sample_key] if samples is None else samples
    seed, k, h = (config[v] for v in ["seed", "k", "h"])
    log_file = "{}_{}_{}_{}_{}.learning_log.txt".format(problem_id, sample_size, seed, k, h)
    log_file_full = os.path.join(results_dir, log_file)
    flats = []
    with open(log_file_full, "r") as f:
        for line in f:
            flats.append(json.loads(line))
    return flats


def get_all_log_messages(results_dir, config, p_id=None, samples=None):
    id_key = "problem_id" if "problem_id" in config else "id"
    sample_key = "sample_size" if "sample_size" in config else "samples"
    problem_id = config[id_key] if p_id is None else p_id
    sample_size = config[sample_key] if samples is None else samples
    seed, k, h = (config[v] for v in ["seed", "k", "h"])

    messages = dict()
    for filename in os.listdir(results_dir):
        pattern = r"{}_{}_{}_(\d+)_(\d+).learning_log.txt".format(problem_id, sample_size, seed)
        match = re.match(pattern, filename)
        if match and (int(match.group(1)) <= k or int(match.group(2)) <= h):
            log_file_full = os.path.join(results_dir, filename)
            flats = []
            with open(log_file_full, "r") as f:
                for line in f:
                    flats.append(json.loads(line))
            messages[(int(match.group(1)), int(match.group(2)))] = flats
    return messages


class TableMaker(object):
    def __init__(self, row_key_type, col_key_type, value_type, delimiter=None):
        self.row_key_type = row_key_type
        self.col_key_type = col_key_type
        self.value_type = value_type
        self.delimiter = "\t" if delimiter is None else delimiter
        self.row_keys = []
        self.col_keys = []
        self.tables = []
        self.indices = dict()

    def extract(self, extraction_type, results_dir, data_dir, config):
        # noinspection PyCallingNonCallable
        return {
            "id": lambda r, d, c: str(c["problem_id"]),
            "index": self.extract_index,
            "name": self.extract_benchmark_name,
            "time": self.extract_time,
            "full_time": self.extract_full_time,
            "time_ratio": lambda r, d, c: self.extract_time(r, d, c) / self.extract_full_time(r, d, c),
            "k": lambda r, d, c: int(c[extraction_type]),
            "h": lambda r, d, c: int(c[extraction_type]),
            "synth_k": self.extract_synth_k,
            "synth_h": self.extract_synth_h,
            "samples": lambda r, d, c: int(c["sample_size"]),
            "l": self.extract_literals,
            "acc": self.extract_accuracy,
            "acc_exact": self.extract_accuracy_exact,
            "rel_acc": self.extract_relative_accuracy,
            "ratio": self.extract_ratio,
            "time_out": self.extract_timeout,
            "active": self.extract_active_set,
            "active_ratio": self.extract_active_ratio,
            "constant": lambda r, d, c: 1,
        }[extraction_type](results_dir, data_dir, config)

    def get_name(self, extraction_type):
        return {
            "id": "Problem ID",
            "index": "Problem index",
            "name": "Name",
            "time": "time (s)",
            "full_time": "Full time (s)",
            "time_ratio": "Ratio time spent on final configuration over total time spent",
            "k": "# terms",
            "h": "# halfspaces",
            "synth_k": "# terms",
            "synth_h": "# halfspaces",
            "samples": "# samples",
            "l": "# literals",
            "acc": "accuracy",
            "acc_exact": "accuracy",
            "rel_acc": "relative accuracy",
            "ratio": "ratio",
            "time_out": "Time-out ratio",
            "active": "Examples used",
            "active_ratio": "ratio of examples used",
            "constant": "Constant",
        }[extraction_type]

    def get_lim(self, extraction_type):
        if extraction_type in ["id", "name", "constant"]:
            return None, None
        elif extraction_type in ["time", "full_time", "active", "index"]:
            return 0, None
        elif extraction_type in ["k", "h", "samples", "l"]:
            return None, None
        elif extraction_type in ["ratio", "time_out", "active_ratio", "time_ratio"]:
            return 0, 1
        elif extraction_type in ["acc", "acc_exact"]:
            return 0.5, 1
        elif extraction_type in ["rel_acc"]:
            return -1, 1
        else:
            raise RuntimeError("Unknown extraction type {}".format(extraction_type))

    def get_x_ticks(self):
        if self.col_key_type in ["k", "h", "l"]:
            return self.col_keys
        else:
            return None

    @staticmethod
    def extract_benchmark_name(results_dir, data_dir, config):
        flat = load()
        return str(flat["lookup"][config["problem_id"]])

    def extract_time(self, results_dir, data_dir, config):
        timed_out = config.get("time_out", False)
        if not timed_out:
            durations = []
            for message in get_log_messages(results_dir, config):
                if message["type"] == "update":
                    durations.append(message["selection_time"] + message["solving_time"])
            return sum(durations)
        else:
            return None  # "({})".format(config["time_limit"])

    def extract_full_time(self, results_dir, data_dir, config):
        timed_out = config.get("time_out", False)
        if not timed_out:
            durations = []
            for key, messages in get_all_log_messages(results_dir, config).items():
                for message in messages:
                    if message["type"] == "update":
                        durations.append(message["selection_time"] + message["solving_time"])
            return sum(durations)
        else:
            return None  # "({})".format(config["time_limit"])

    def extract_literals(self, results_dir, data_dir, config):
        with open(os.path.join(data_dir, "{}.txt".format(str(config["problem_id"])))) as f:
            s_problem = generator.import_synthetic_data(json.load(f))
        return s_problem.synthetic_problem.literals

    def extract_synth_k(self, results_dir, data_dir, config):
        with open(os.path.join(data_dir, "{}.txt".format(str(config["problem_id"])))) as f:
            s_problem = generator.import_synthetic_data(json.load(f))
        return (s_problem.synthetic_problem.k - self.extract("k", results_dir, data_dir, config)) / s_problem.synthetic_problem.k

    def extract_synth_h(self, results_dir, data_dir, config):
        with open(os.path.join(data_dir, "{}.txt".format(str(config["problem_id"])))) as f:
            s_problem = generator.import_synthetic_data(json.load(f))
        return (s_problem.synthetic_problem.h - self.extract("h", results_dir, data_dir, config)) / s_problem.synthetic_problem.h

    def extract_accuracy(self, results_dir, data_dir, config):
        timed_out = config.get("time_out", False)
        if not timed_out:
            return config["approx_accuracy"]["1000"][0]["acc"]
        else:
            return None

    def extract_accuracy_exact(self, results_dir, data_dir, config):
        timed_out = config.get("time_out", False)
        if not timed_out:
            return config["exact_accuracy"]
        else:
            return None

    def extract_relative_accuracy(self, results_dir, data_dir, config):
        return self.extract_accuracy(results_dir, data_dir, config) - self.extract_ratio(results_dir, data_dir, config)

    def extract_ratio(self, results_dir, data_dir, config):
        return config["approx_ratio"]["1000"][0]["ratio"]

    def extract_timeout(self, results_dir, data_dir, config):
        return 1 if config.get("time_out", False) else 0

    def extract_active_set(self, results_dir, data_dir, config):
        timed_out = config.get("time_out", False)
        if not timed_out:
            examples = 20
            for message in get_log_messages(results_dir, config):
                if message["type"] == "update":
                    examples += min(10, len(message["indices"]))
            return examples
        else:
            return None  # "({})".format(config["time_limit"])

    def extract_active_ratio(self, results_dir, data_dir, config):
        timed_out = config.get("time_out", False)
        if not timed_out:
            return self.extract_active_set(results_dir, data_dir, config) / int(config["sample_size"])
        else:
            return None  # "({})".format(config["time_limit"])

    def extract_index(self, results_dir, data_dir, config):
        p_id = str(config["problem_id"])
        if p_id not in self.indices:
            self.indices[p_id] = len(self.indices)
        return self.indices[p_id]

    def load_table(self, results_dir, data_dir, name=None):
        problems = load_results(results_dir)

        unique_row_keys = set()
        unique_col_keys = set()
        table = dict()

        table["name"] = name

        for problem_id in problems:
            for sample_size in problems[problem_id]:
                config = problems[problem_id][sample_size]
                config["problem_id"] = problem_id
                config["sample_size"] = sample_size

                row_key = self.extract(self.row_key_type, results_dir, data_dir, config)
                col_key = self.extract(self.col_key_type, results_dir, data_dir, config)
                value = self.extract(self.value_type, results_dir, data_dir, config)

                unique_row_keys.add(row_key)
                unique_col_keys.add(col_key)
                if (row_key, col_key) not in table:
                    table[(row_key, col_key)] = []
                table[(row_key, col_key)].append(value)

        unique_row_keys = unique_row_keys | set(self.row_keys)
        unique_col_keys = unique_col_keys | set(self.col_keys)
        row_key_is_int = all(isinstance(k, (int, long)) for k in unique_row_keys)
        self.row_keys = list(sorted(unique_row_keys, key=(int if row_key_is_int else str)))
        col_key_is_int = all(isinstance(k, (int, long)) for k in unique_col_keys)
        self.col_keys = list(sorted(unique_col_keys, key=(int if col_key_is_int else str)))

        self.tables.append(table)

    def to_txt(self, i, aggregate=False):
        import numpy

        table = self.tables[i]

        def get_val(_key):
            value = None
            if _key not in table:
                value = None
            elif len(table[_key]) == 1:
                value = table[_key][0]
            else:
                initial = table[_key][0]
                for i in range(1, len(table[_key])):
                    initial += table[_key][i]
                value = initial / len(table[_key])
            return value if value is not None else numpy.nan

        lines = [self.delimiter.join([""] + [str(k) for k in self.col_keys])]
        if aggregate:
            name = "Average " + self.get_name(self.value_type)
            averages = [numpy.nanmean(numpy.array([get_val((rk, ck)) for rk in self.row_keys])) for ck in self.col_keys]
            deviation = [numpy.nanstd(numpy.array([get_val((rk, ck)) for rk in self.row_keys])) for ck in self.col_keys]
            lines.append(self.delimiter.join([str(name)] + ["{} +/- {}".format(a, d) for a, d in zip(averages, deviation)]))
        else:
            for rk in self.row_keys:
                lines.append(self.delimiter.join([str(rk)] + [str(get_val((rk, ck))) for ck in self.col_keys]))
        return "\n".join(lines)

    def plot_table(self, filename=None, index=None, y_min=None, y_max=None, x_min=None, x_max=None, legend_pos=None):
        import rendering
        import numpy

        def get_val(_table, _key):
            if _key not in _table:
                return numpy.nan
            else:
                return numpy.nanmean([v if v is not None else numpy.nan for v in _table[_key]])

        scatter = rendering.ScatterData("", numpy.array(self.col_keys))
        y_lim = self.get_lim(self.value_type)
        if y_min is not None:
            y_lim[0] = y_min
        if y_max is not None:
            y_lim[1] = y_max
        x_lim = self.get_lim(self.col_key_type)
        if x_min is not None:
            x_lim[0] = x_min
        if x_max is not None:
            x_lim[1] = x_max

        scatter.y_lim(y_lim)
        scatter.x_lim(x_lim)
        series = [
            [numpy.array([get_val(table, (rk, ck)) for ck in self.col_keys]) for rk in self.row_keys]
            for table in self.tables
        ]
        if index is None:
            for i in range(len(self.tables)):
                series_array = numpy.array(series[i])
                legend_name = "Average " + self.get_name(self.value_type)
                if "name" in self.tables[i]:
                    legend_name += " " + self.tables[i]["name"]
                scatter.add_data(legend_name, numpy.nanmean(series_array, 0), numpy.nanstd(numpy.array(series[i]), 0))
            # std_dev_series =
        else:
            for i in range(len(self.row_keys)):
                scatter.add_data(self.row_keys[i], series[index][i])

        label_y = self.get_name(self.value_type).capitalize()
        label_x = self.get_name(self.col_key_type).capitalize()
        scatter.plot(filename=filename, size=(4, 2.8), lines=True, log_x=False, log_y=False, label_y=label_y, label_x=label_x,
                     x_ticks=self.get_x_ticks(), legend_pos=legend_pos)



import argparse

import os

import math


if __name__ == "__main__":
    def parse_args():
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="mode")

        scan_parser = subparsers.add_parser("scan", help="Scan the directory and load smt problems")
        scan_parser.add_argument("-d", "--dir", default=None, help="Specify the directory to load files from")

        learn_parser = subparsers.add_parser("learn", help="Learn SMT formulas")
        learn_parser.add_argument("dir", help="Specify the results directory")
        learn_parser.add_argument("-s", "--samples", type=int, help="Specify the number of samples for learning")
        learn_parser.add_argument("-a", "--all", default=False, action="store_true",
                                  help="If set, learning will not use incremental mode and include all examples")
        learn_parser.add_argument("-d", "--dnf", default=False, action="store_true",
                                  help="If set, learning bias is DNF instead of CNF")

        table_parser = subparsers.add_parser("table", help="Types can be: [time, k, h, id, acc, samples, l]")
        table_parser.add_argument("row_key", help="Specify the row key type")
        table_parser.add_argument("col_key", default=None, help="Specify the col key type")
        table_parser.add_argument("value", default=None, help="Specify the value type")
        table_parser.add_argument("dirs", nargs="*", help="Specify the directories to load files from, always in pairs:"
                                                          "result_dir, data_dir")

        table_subparsers = table_parser.add_subparsers(dest="command")
        table_print_parser = table_subparsers.add_parser("print", help="Print the table")
        table_print_parser.add_argument("-d", "--delimiter", default="\t", help="Specify the delimiter (default=tab)")
        table_print_parser.add_argument("-a", "--aggregate", default=False, action="store_true",
                                       help="Aggregate the rows in the plot")

        table_plot_parser = table_subparsers.add_parser("plot", help="Plot the table")
        table_plot_parser.add_argument("-a", "--aggregate", default=False, action="store_true",
                                       help="Aggregate the rows in the plot")
        table_plot_parser.add_argument("--y_min", default=None, type=float, help="Minimum value for y")
        table_plot_parser.add_argument("--y_max", default=None, type=float, help="Maximum value for y")
        table_plot_parser.add_argument("--x_min", default=None, type=float, help="Minimum value for x")
        table_plot_parser.add_argument("--x_max", default=None, type=float, help="Maximum value for x")
        table_plot_parser.add_argument("--legend_pos", default=None, type=str, help="Legend position")
        table_plot_parser.add_argument("-o", "--output", default=None, help="Specify the output file")

        combine_parser = subparsers.add_parser("combine", help="Combine multiple results directories")
        combine_parser.add_argument("output_dir", help="The output directory to summarize results in")
        combine_parser.add_argument("input_dirs", nargs="*", help="Specify the directories to combine")
        combine_parser.add_argument("-b", "--bias", default=None, help="Specify the bias")
        combine_parser.add_argument("-p", "--prefix", default=None, help="Specify the prefix for input dirs")

        gen_parser = subparsers.add_parser("generate", help="Generate synthetic examples")
        gen_parser.add_argument("data_dir")
        gen_parser.add_argument("-n", "--data_sets", default=10, type=int)
        gen_parser.add_argument("--prefix", default="synthetics")
        gen_parser.add_argument("-b", "--bool_count", default=2, type=int)
        gen_parser.add_argument("-r", "--real_count", default=2, type=int)
        gen_parser.add_argument("--bias", default="cnf")
        gen_parser.add_argument("-k", "--k", default=3, type=int)
        gen_parser.add_argument("-l", "--literals", default=4, type=int)
        gen_parser.add_argument("--half_spaces", default=7, type=int)
        gen_parser.add_argument("-s", "--samples", default=1000, type=int)
        gen_parser.add_argument("--ratio", default=90, type=int)
        gen_parser.add_argument("-p", "--plot_dir", default=None)
        gen_parser.add_argument("-e", "--errors", default=0, type=int)

        migration_parser = subparsers.add_parser("migrate", help="Migrate files to newer or extended versions")
        migration_subparsers = migration_parser.add_subparsers(dest="type")

        migration_fix_parser = migration_subparsers.add_parser("fix", help="Fix result files")
        migration_fix_parser.add_argument("results_dir", help="Specify the result directory")
        migration_fix_parser.add_argument("-b", "--bias", default=None, help="Specify the bias")

        migration_acc_parser = migration_subparsers.add_parser("accuracy", help="Add accuracy to result files")
        migration_acc_parser.add_argument("results_dir", help="Specify the result directory")
        migration_acc_parser.add_argument("-d", "--data_dir", help="Specify the data directory for synthetic problems")
        migration_acc_parser.add_argument("-s", "--samples", default=None, help="Specify the number of samples", type=int)
        migration_acc_parser.add_argument("-f", "--force", default=False, action="store_true", help="Overwrites existing values")

        migration_ratio_parser = migration_subparsers.add_parser("ratio", help="Add ratio to result files")
        migration_ratio_parser.add_argument("results_dir", help="Specify the result directory")
        migration_ratio_parser.add_argument("-d", "--data_dir", help="Specify the data directory for synthetic problems")
        migration_ratio_parser.add_argument("-s", "--samples", default=None, help="Specify the number of samples", type=int)
        migration_ratio_parser.add_argument("-f", "--force", default=False, action="store_true", help="Overwrites existing values")

        args = parser.parse_args()

        if args.mode == "scan":
            full_dir = os.path.abspath(args.filename)
            root_dir = os.path.dirname(full_dir)

            import smt_scan
            smt_scan.scan(full_dir, root_dir)
            smt_scan.analyze(root_dir)
            smt_scan.ratios()
        elif args.mode == "learn":
            import smt_scan
            smt_scan.learn(args.samples, args.dir, args.all, args.dnf)
        elif args.mode == "table":
            import smt_scan
            table = smt_scan.TableMaker(args.row_key, args.col_key, args.value)
            for i in range(int(math.floor(len(args.dirs) / 3))):
                table.load_table(args.dirs[3 * i], args.dirs[3 * i + 1], args.dirs[3 * i + 2])
            if args.command == "print":
                table.delimiter = args.delimiter
                print(table.to_txt(0, args.aggregate))
            elif args.command == "plot":
                table.plot_table(args.output, None if args.aggregate else 0, args.y_min, args.y_max, args.x_min, args.x_max, args.legend_pos)
            else:
                print("Error: unknown table command {}".format(args.command))
        elif args.mode == "combine":
            import combine_results
            combine_results.combine(args.output_dir, args.input_dirs, args.bias, args.prefix)
        elif args.mode == "generate":
            from generator import generate_random
            generate_random(args.data_sets, args.prefix, args.bool_count, args.real_count, args.bias, args.k,
                            args.literals, args.half_spaces, args.samples, args.ratio, args.errors, args.data_dir,
                            args.plot_dir)
        elif args.mode == "migrate":
            import migrate
            if args.type == "fix":
                migrate.migrate_results(args.results_dir, args.bias)
            elif args.type == "accuracy":
                migrate.add_accuracy(args.results_dir, args.data_dir, args.samples, args.force)
            elif args.type == "ratio":
                migrate.add_ratio(args.results_dir, args.data_dir, args.samples, args.force)
            else:
                print("Error: unknown migration type {}".format(args.type))
        else:
            print("Error: unknown mode {}".format(args.mode))


    parse_args()
